'''
HCE project, Python bindings, Distributed Crawler application.
BatchTasksManager object and related classes definitions.

@package: dc
@author bgv bgv.hce@gmail.com
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
'''


import time
import logging
# import pickle
try:
  import cPickle as pickle
except ImportError:
  import pickle

from app.BaseServerManager import BaseServerManager
from app.Utils import SQLExpression, varDump
from dc import EventObjects
import dtm.EventObjects
from transport.ConnectionBuilderLight import ConnectionBuilderLight
import dc.Constants as DC_CONSTS
import dtm.Constants as DTM_CONSTS
import transport.Consts as TRANSPORT_CONSTS
import app.Utils as Utils  # pylint: disable=F0401


# Logger initialization
logger = logging.getLogger(DC_CONSTS.LOGGER_NAME)


# #The BatchTasksManager class, is a main crawling logic of DC application.
#
# This object is a main crawling batches algorithms of DC application.
# It supports complete DTM protocol requests and process responses from DTM server, operates with tasks and monitors
# tasks state.
#
class BatchTasksManager(BaseServerManager):

  DTM_TASK_CHECK_STATE_METHOD_STATUS = 0
  DTM_TASK_CHECK_STATE_METHOD_STATE = 1

  # Configuration settings options names
  CONFIG_SERVER = "server"
  CONFIG_DTMD_HOST = "DTMDHost"
  CONFIG_DTMD_PORT = "DTMDPort"
  CONFIG_DTMD_TIMEOUT = "DTMDTimeout"
  CONFIG_SITES_MANAGER_CLIENT = "clientSitesManager"
  CONFIG_POLLING_TIMEOUT = "PollingTimeout"
  CONFIG_DRCE_CRAWLER_APP_NAME = "DRCECrawlerAppName"
  CONFIG_BATCH_DEFAULT_MAX_TIME = "BatchDefaultMaxExecutionTime"
  CONFIG_MAX_URLS_BATCH = "BatchDefaultMaxURLs"
  CONFIG_ORDER_BY_URLS_BATCH = "BatchDefaultOrderByURLs"
  CONFIG_RET_URLS_MAX_NUMBER = "ReturnURLsMaxNumber"
  CONFIG_RET_URLS_PERIOD = "ReturnURLsPeriod"
  CONFIG_RET_URLS_TTL = "ReturnURLsTTL"

  CONFIG_INCR_MIN_FREQ = "IncrMinFreq"
  CONFIG_INCR_MAX_DEPTH = "IncrMaxDepth"
  CONFIG_INCR_MAX_URL = "IncrMaxURLs"
  CONFIG_INCR_CRAWL_PERIOD = "IncrPeriod"

  CONFIG_CRAWLED_URLS_STRATEGY = "CrawledURLStrategy"
  CONFIG_MAX_TASKS_BATCH = "BatchDefaultMaxTasks"

  CONFIG_REGULAR_CRAWL_PERIOD = "RegularCrawlingPeriod"
  CONFIG_BATCH_QUEUE_PERIOD = "BatchQueueProcessingPeriod"
  CONFIG_BATCH_QUEUE_TASK_TTL = "BatchQueueTaskTTL"

  CONFIG_BATCH_QUEUE_TASK_CHECK_METHOD = "BatchQueueTaskCheckStateMethod"

  # The tasks's strategy configuration parameters for DTM service
  CONFIG_BATCH_DEFAULT_STRATEGY_IO_WAIT_MAX = "BatchTask_IO_WAIT_MAX"
  CONFIG_BATCH_DEFAULT_STRATEGY_CPU_LOAD_MAX = "BatchTask_CPU_LOAD_MAX"
  CONFIG_BATCH_DEFAULT_STRATEGY_STRATEGY_RDELAY = "BatchTask_RDELAY"
  CONFIG_BATCH_DEFAULT_STRATEGY_RETRY = "BatchTask_RETRY"


  # #constructor
  # initialize fields
  #
  # @param configParser config parser object
  # @param connectBuilderLight connection builder light
  #
  def __init__(self, configParser, connectionBuilderLight=None):
    super(BatchTasksManager, self).__init__()

    # Instantiate the connection builder light if not set
    if connectionBuilderLight == None:
      connectionBuilderLight = ConnectionBuilderLight()

    # Batches counter init in stat vars
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_TOTAL_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Batches in queue counter init in stat vars
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_QUEUE_NAME, 0, self.STAT_FIELDS_OPERATION_SET)
    # Batches that fault processing counter init in stat vars
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_FAULT_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Batches that not empty counter init in stat vars
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_FILLED_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Batches urls total counter init in stat vars
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URLS_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Fault batches urls total counter init in stat vars
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URLS_FAULT_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Crawling URL_FETCH requests counter name for stat variables
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URL_FETCH_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Crawling cancelled URL_FETCH requests counter name for stat variables
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URL_FETCH_CANCELLED_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Crawling delete task requests fault counter name for stat variables
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_DELETE_FAULT_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Crawling check task requests fault counter name for stat variables
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_CHECK_FAULT_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Crawling batches urls returned counter name for stat variables
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URLS_RET_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Crawling incremental URL_FETCH requests counter name for stat variables
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URL_FETCH_INCR_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Crawling batches fault TTL counter name for stat variables
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_FAULT_TTL_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)

    # Get configuration settings
    className = self.__class__.__name__
    self.serverName = configParser.get(className, self.CONFIG_SERVER)
    self.clientSitesManagerName = configParser.get(className, self.CONFIG_SITES_MANAGER_CLIENT)
    # Configuration settings for DTMD server interaction
    self.configVars[self.CONFIG_DTMD_HOST] = configParser.get(className, self.CONFIG_DTMD_HOST)
    self.configVars[self.CONFIG_DTMD_PORT] = configParser.get(className, self.CONFIG_DTMD_PORT)
    self.configVars[self.CONFIG_DTMD_TIMEOUT] = configParser.getint(className, self.CONFIG_DTMD_TIMEOUT)

    # Max URLs per batch
    self.configVars[self.CONFIG_MAX_URLS_BATCH] = configParser.getint(className, self.CONFIG_MAX_URLS_BATCH)

    # Set connections poll timeout, defines period of HCE cluster monitoring cycle, msec
    self.configVars[self.POLL_TIMEOUT_CONFIG_VAR_NAME] = configParser.getint(className, self.CONFIG_POLLING_TIMEOUT)
    # Configuration settings for incremental crawling
    self.configVars[DC_CONSTS.INCR_MIN_FREQ_CONFIG_VAR_NAME] = configParser.getint(className, self.CONFIG_INCR_MIN_FREQ)
    self.configVars[DC_CONSTS.INCR_MAX_DEPTH_CONFIG_VAR_NAME] = configParser.getint(className,
                                                                                    self.CONFIG_INCR_MAX_DEPTH)
    self.configVars[DC_CONSTS.INCR_MAX_URLS_CONFIG_VAR_NAME] = configParser.getint(className, self.CONFIG_INCR_MAX_URL)

    # Set crawler task app name
    self.configVars[self.CONFIG_DRCE_CRAWLER_APP_NAME] = configParser.get(className, self.CONFIG_DRCE_CRAWLER_APP_NAME)
    self.configVars[self.CONFIG_BATCH_DEFAULT_MAX_TIME] = \
      configParser.getint(className, self.CONFIG_BATCH_DEFAULT_MAX_TIME)

    # Create connections and raise bind or connect actions for correspondent connection type
    serverConnection = connectionBuilderLight.build(TRANSPORT_CONSTS.SERVER_CONNECT, self.serverName)
    sitesManagerConnection = connectionBuilderLight.build(TRANSPORT_CONSTS.CLIENT_CONNECT, self.clientSitesManagerName)

    # Init the DTMD connection
    self.dtmdConnection = connectionBuilderLight.build(TRANSPORT_CONSTS.CLIENT_CONNECT,
                                                       self.configVars[self.CONFIG_DTMD_HOST] + ":" + \
                                                       self.configVars[self.CONFIG_DTMD_PORT],
                                                       TRANSPORT_CONSTS.TCP_TYPE)

    # Add connections to the polling set
    self.addConnection(self.serverName, serverConnection)
    self.addConnection(self.clientSitesManagerName, sitesManagerConnection)

    # Set event handler for URL_FETCH_RESPONSE event
    self.setEventHandler(DC_CONSTS.EVENT_TYPES.URL_FETCH_RESPONSE, self.onURLFetchResponse)
    # Set event handler for URL_UPDATE_RESPONSE event
    self.setEventHandler(DC_CONSTS.EVENT_TYPES.URL_UPDATE_RESPONSE, self.onURLUpdateResponse)
    # Set event handler for URL_DELETE_RESPONSE event
    self.setEventHandler(DC_CONSTS.EVENT_TYPES.URL_DELETE_RESPONSE, self.onURLDeleteResponse)

    # Initialize the DTM tasks queue, the key is taskId, the value is the Batch object
    self.dtmTasksQueue = {}

    # Init config vars storage for return URLs selected for processing
    self.configVars[self.CONFIG_RET_URLS_MAX_NUMBER] = configParser.getint(className, self.CONFIG_RET_URLS_MAX_NUMBER)
    self.configVars[self.CONFIG_RET_URLS_PERIOD] = configParser.getint(className, self.CONFIG_RET_URLS_PERIOD)
    self.configVars[self.CONFIG_RET_URLS_TTL] = configParser.getint(className, self.CONFIG_RET_URLS_TTL)
    self.processSelectedURLsRetLastTs = time.time()

    # Incremental crawling init
    self.configVars[self.CONFIG_INCR_CRAWL_PERIOD] = configParser.getint(className, self.CONFIG_INCR_CRAWL_PERIOD)
    self.processIncrCrawlLastTs = time.time()

    # Crawled URLs strategy for batch
    self.configVars[self.CONFIG_CRAWLED_URLS_STRATEGY] = configParser.getint(className,
                                                                             self.CONFIG_CRAWLED_URLS_STRATEGY)
    #Batch default order by criterion to fetch URLs
    self.configVars[self.CONFIG_ORDER_BY_URLS_BATCH] = configParser.get(className, self.CONFIG_ORDER_BY_URLS_BATCH)

    #Max tasks in batch queue, if limit reached new batch tasks will not be started; zero means unlimited
    self.configVars[self.CONFIG_MAX_TASKS_BATCH] = configParser.getint(className, self.CONFIG_MAX_TASKS_BATCH)

    # Regular crawling init
    self.configVars[self.CONFIG_REGULAR_CRAWL_PERIOD] = configParser.getint(className, self.CONFIG_REGULAR_CRAWL_PERIOD)
    self.processRegularCrawlLastTs = time.time()

    # The batch queue processing init
    self.configVars[self.CONFIG_BATCH_QUEUE_PERIOD] = configParser.getint(className, self.CONFIG_BATCH_QUEUE_PERIOD)
    self.processBatchQueuelLastTs = time.time()

    # The batch queue task TTL, sec
    self.configVars[self.CONFIG_BATCH_QUEUE_TASK_TTL] = configParser.getint(className, self.CONFIG_BATCH_QUEUE_TASK_TTL)

    # The batch queue tasks state check method, see ini file comments
    self.configVars[self.CONFIG_BATCH_QUEUE_TASK_CHECK_METHOD] = \
      configParser.getint(className, self.CONFIG_BATCH_QUEUE_TASK_CHECK_METHOD)

  # The tasks's strategy configuration parameters for DTM service load
    self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_IO_WAIT_MAX] = \
      configParser.getint(className, self.CONFIG_BATCH_DEFAULT_STRATEGY_IO_WAIT_MAX)
    self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_CPU_LOAD_MAX] = \
      configParser.getint(className, self.CONFIG_BATCH_DEFAULT_STRATEGY_CPU_LOAD_MAX)
    self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_STRATEGY_RDELAY] = \
      configParser.getint(className, self.CONFIG_BATCH_DEFAULT_STRATEGY_STRATEGY_RDELAY)
    self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_RETRY] = \
      configParser.getint(className, self.CONFIG_BATCH_DEFAULT_STRATEGY_RETRY)



  # #Events wait timeout handler, for timeout state of the connections polling. Executes periodical check of DTM tasks
  # and initiate the main crawling iteration cycle
  #
  def on_poll_timeout(self):
    logger.debug("Periodic iteration started.")
    try:
      if time.time() - self.processRegularCrawlLastTs > self.configVars[self.CONFIG_REGULAR_CRAWL_PERIOD]:
        self.processRegularCrawlLastTs = time.time()
        logger.debug("URL_FETCH for regular crawling iteration!")
        # Process the first step of crawling iteration
        if self.configVars[self.CONFIG_MAX_TASKS_BATCH] == 0 or\
          (self.configVars[self.CONFIG_MAX_TASKS_BATCH] > 0 and\
           self.configVars[self.CONFIG_MAX_TASKS_BATCH] > len(self.dtmTasksQueue)):
          self.sendURLFetchRequest()
          self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URL_FETCH_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
        else:
          self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URL_FETCH_CANCELLED_NAME, 1,
                               self.STAT_FIELDS_OPERATION_ADD)
          logger.debug("Max batch tasks %s in queue reached, new batch is not created!",
                       str(self.configVars[self.CONFIG_MAX_TASKS_BATCH]))

      if time.time() - self.processBatchQueuelLastTs > self.configVars[self.CONFIG_BATCH_QUEUE_PERIOD]:
        self.processBatchQueuelLastTs = time.time()
        logger.debug("Process DTM tasks queue!")
        # Process the DTM tasks queue
        self.processDTMTasksQueue()

      # Process periodic selected for crawling URLs return
      if time.time() - self.processSelectedURLsRetLastTs > self.configVars[self.CONFIG_RET_URLS_PERIOD]:
        logger.debug("Now time to perform URLs return, interval " + str(self.configVars[self.CONFIG_RET_URLS_PERIOD]))
        self.processSelectedURLsRetLastTs = time.time()
        self.processSelectedURLsReturn()

      if time.time() - self.processIncrCrawlLastTs > self.configVars[self.CONFIG_INCR_CRAWL_PERIOD]:
        logger.debug("URL_FETCH for incremental crawling iteration!")
        self.processIncrCrawlLastTs = time.time()
        # Process the first step of incremental crawling iteration
        self.sendIncrURLRequest()

    except Exception as err:
      logger.error("Exception: " + str(err.message) + "\n" + Utils.getTracebackInfo())

    logger.debug("Periodic iteration finished.")



  # #onURLFetchResponse event handler
  #
  # @param event instance of Event object
  def onURLFetchResponse(self, event):
    try:
      logger.debug("Reply received on URL fetch: %s", varDump(event))
      if event.cookie is not None and isinstance(event.cookie, dict):
        crawlerType = event.cookie.get('type', EventObjects.Batch.TYPE_NORMAL_CRAWLER)
      else:
        crawlerType = EventObjects.Batch.TYPE_NORMAL_CRAWLER
      clientResponse = event.eventObj
      if clientResponse.errorCode == EventObjects.ClientResponse.STATUS_OK:
        if len(clientResponse.itemsList) > 0:
          if event.cookie is not None and\
            (isinstance(event.cookie, dict) and\
             EventObjects.Batch.OPERATION_TYPE_NAME in event.cookie and\
             (event.cookie[EventObjects.Batch.OPERATION_TYPE_NAME] == EventObjects.Batch.TYPE_NORMAL_CRAWLER or\
              event.cookie[EventObjects.Batch.OPERATION_TYPE_NAME] == EventObjects.Batch.TYPE_INCR_CRAWLER)):
            # Process response for main crawling URLFetch request
            batch = self.makeBatchFromClientResponseItems(clientResponse.itemsList, crawlerType)
            if len(batch.items) > 0:
              taskId = self.sendBatchTaskToDTM(batch)
              if taskId > 0:
                logger.debug("DTM batch was set, taskId=%s", str(taskId))
                # Insert the Batch object in to the queue
                batch.QueuedTs = time.time()
                self.dtmTasksQueue[taskId] = batch
                self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_QUEUE_NAME, len(self.dtmTasksQueue),
                                     self.STAT_FIELDS_OPERATION_SET)
              else:
                logger.error("Error send the Batch object to DTM!")
                if crawlerType == EventObjects.Batch.TYPE_NORMAL_CRAWLER:
                  # Update URLs state back to New to get possibility to select them next time
                  self.sendURLUpdate(batch.items, batch.id, False)
            else:
              logger.debug("There is no items in batch!")
          else:
            # Process response for URLs return URLFetch request
            urls = self.getURLsCountFromClientResponseItems(clientResponse.itemsList)
            self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URLS_RET_NAME, urls, self.STAT_FIELDS_OPERATION_ADD)
            logger.debug(str(urls) + " URLs returned back to state NEW because processing TTL exceed.")
        else:
          logger.debug("There is empty clientResponse.itemsList")
      else:
        logger.error("URLFetch response error: " + str(clientResponse.errorCode) + " : " + clientResponse.errorMessage)
    except Exception as err:
      logger.error("Exception: " + str(err.message) + "\n" + Utils.getTracebackInfo())



  # #onURLUpdateResponse event handler
  #
  # @param event instance of Event object
  def onURLUpdateResponse(self, event):
    try:
      logger.debug("Reply received on URL update.")
      clientResponse = event.eventObj
      if clientResponse.errorCode == EventObjects.ClientResponse.STATUS_OK:
        if len(clientResponse.itemsList) > 0:
          for clientResponseItem in clientResponse.itemsList:
            if clientResponseItem.errorCode != EventObjects.ClientResponseItem.STATUS_OK:
              logger.error("URLUpdate response error: " + str(clientResponseItem.errorCode) + " : " + \
                           clientResponseItem.errorMessage + ", host:" + clientResponseItem.host + ", port:" + \
                           clientResponseItem.port + ", node:" + clientResponseItem.node + "!")
        else:
          logger.error("URLUpdate response empty list!")
      else:
        logger.error("URLUpdate response error:" + str(clientResponse.errorCode) + " : " + clientResponse.errorMessage)
    except Exception as err:
      logger.error("Exception: " + str(err.message) + "\n" + Utils.getTracebackInfo())



  # #make the Batch object from the ClientResponse object items
  # @param clientResponseItems The list of ClientResponseItem objects
  #
  def sendBatchTaskToDTM(self, batch):
    taskId = 0

    # Prepare NewTask object
    newTaskObj = dtm.EventObjects.NewTask(self.configVars[self.CONFIG_DRCE_CRAWLER_APP_NAME])
    newTaskObj.setSessionVar("tmode", dtm.EventObjects.Task.TASK_MODE_ASYNCH)
    newTaskObj.setSessionVar("shell", "hce-starter-dc.sh")
    newTaskObj.setSessionVar("time_max", int(self.configVars[self.CONFIG_BATCH_DEFAULT_MAX_TIME]) * 1000)
    # Configure task's strategy
    if self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_IO_WAIT_MAX] > 0:
      newTaskObj.setStrategyVar(dtm.EventObjects.Task.STRATEGY_IO_WAIT_MAX,
                                self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_IO_WAIT_MAX])
    if self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_CPU_LOAD_MAX] > 0:
      newTaskObj.setStrategyVar(dtm.EventObjects.Task.STRATEGY_CPU_LOAD_MAX,
                                self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_CPU_LOAD_MAX])
    if self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_STRATEGY_RDELAY] > 0:
      newTaskObj.setStrategyVar(dtm.EventObjects.Task.STRATEGY_RDELAY,
                                self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_STRATEGY_RDELAY])
    if self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_RETRY] > 0:
      newTaskObj.setStrategyVar(dtm.EventObjects.Task.STRATEGY_RETRY,
                                self.configVars[self.CONFIG_BATCH_DEFAULT_STRATEGY_RETRY])

    batch.id = newTaskObj.id
    newTaskObj.input = pickle.dumps(batch)
    newTaskEvent = self.eventBuilder.build(DTM_CONSTS.EVENT_TYPES.NEW_TASK, newTaskObj)
    generalResponse = self.dtmdRequestExecute(newTaskEvent, self.configVars[self.CONFIG_DTMD_TIMEOUT])
    if generalResponse is not None:
      if generalResponse.errorCode == dtm.EventObjects.GeneralResponse.ERROR_OK:
        # New crawling Batch task set successfully
        taskId = newTaskObj.id
      else:
        # Some error of task operation
        logger.error("DTM set batch task error: " + str(generalResponse.errorCode) + " : " + \
                     generalResponse.errorMessage + ", statuses:" + varDump(generalResponse))
    else:
      logger.error("DTM set batch task response error, possible timeout!")

    #TODO: return the Task Id any case error or not to check it state later
    taskId = newTaskObj.id

    return taskId



  # #make the Batch object from the ClientResponse object items

  # @param clientResponseItems The list of ClientResponseItem objects
  # @return the Batch object instance
  def makeBatchFromClientResponseItems(self, clientResponseItems, crawlerType):
    batch = EventObjects.Batch(0, crawlerType=crawlerType)
    batchItemsCounter = 0
    batchItemsTotalCounter = 0
    uniqueURLsCRCDic = {}
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_TOTAL_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)

    for item in clientResponseItems:
      if item.errorCode == EventObjects.ClientResponseItem.STATUS_OK:
        if type(item.itemObject) == type(list()):
          for url in item.itemObject:
            batchItemsTotalCounter = batchItemsTotalCounter + 1
            if type(url) == type(EventObjects.URL("", "")):
              batchItem = EventObjects.BatchItem(url.siteId, url.urlMd5, url)
              itemId = str(url.siteId) + ":" + str(url.urlMd5)
              if itemId not in uniqueURLsCRCDic:
                uniqueURLsCRCDic[itemId] = batchItem
                logger.debug("Insert batchItem: %s", varDump(batchItem))
                batch.items.append(batchItem)
                batchItemsCounter = batchItemsCounter + 1
                self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URLS_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
            else:
              logger.error("Wrong object type in the itemObject.item: " + str(type(url)) + \
                           " but 'URL' expected")
          if batchItemsCounter > 0:
            self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_FILLED_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
        else:
          logger.error("Wrong object type in the ClientResponseItem.itemObject: " + str(type(item.itemObject)) + \
                       " but 'list' expected")
      else:
        logger.debug("ClientResponseItem error: " + str(item.errorCode) + " : " + item.errorMessage)

    logger.debug("Batch object created, items: " + str(batchItemsTotalCounter) + " total, " + str(batchItemsCounter) + \
                " added!")

    return batch



  # #Enumerate URLs number in the ClientResponse object items
  #
  # @param clientResponseItems The list of ClientResponseItem objects
  # @return the unique URLs number
  def getURLsCountFromClientResponseItems(self, clientResponseItems, unique=True):
    batchItemsCounter = 0
    batchItemsTotalCounter = 0
    uniqueURLsCRCDic = {}

    for item in clientResponseItems:
      if item.errorCode == EventObjects.ClientResponseItem.STATUS_OK:
        if type(item.itemObject) == type(list()):
          for url in item.itemObject:
            batchItemsTotalCounter = batchItemsTotalCounter + 1
            if type(url) == type(EventObjects.URL("", "")):
              itemId = str(url.siteId) + ":" + str(url.urlMd5)
              if itemId not in uniqueURLsCRCDic:
                uniqueURLsCRCDic[itemId] = 1
                batchItemsCounter = batchItemsCounter + 1
              else:
                uniqueURLsCRCDic[itemId] += 1
            else:
              logger.error("Wrong object type in the itemObject.item: " + str(type(url)) + \
                           " but 'URL' expected")
        else:
          logger.error("Wrong object type in the ClientResponseItem.itemObject: " + str(type(item.itemObject)) + \
                       " but 'list' expected")
      else:
        logger.debug("ClientResponseItem error: " + str(item.errorCode) + " : " + item.errorMessage)

    logger.debug("Unique URLs: " + str(batchItemsCounter) + ", total URLs: " + str(batchItemsTotalCounter))

    if unique:
      return batchItemsCounter
    else:
      return batchItemsTotalCounter



  # #Send URLFetch request
  #
  #
  def sendURLFetchRequest(self):
    # Process the first step of crawling iteration
    urlUpdate = EventObjects.URLUpdate(0, "", EventObjects.URLStatus.URL_TYPE_MD5, None,
                                       EventObjects.URL.STATUS_SELECTED_CRAWLING)
    urlUpdate.tcDate = SQLExpression("NOW()")
    # Create URLFetch object with URLUpdate to update selected URLs state
    sitesCriterions = {EventObjects.URLFetch.CRITERION_WHERE: "`State`=" + str(EventObjects.Site.STATE_ACTIVE)}
    urlCriterions = {EventObjects.URLFetch.CRITERION_WHERE: "`Status`=" + str(EventObjects.URL.STATUS_NEW),
                     EventObjects.URLFetch.CRITERION_ORDER: self.configVars[self.CONFIG_ORDER_BY_URLS_BATCH],
                     EventObjects.URLFetch.CRITERION_LIMIT: str(self.configVars[self.CONFIG_MAX_URLS_BATCH])}
    urlFetch = EventObjects.URLFetch(None, urlCriterions, sitesCriterions, urlUpdate)
    urlFetch.algorithm = EventObjects.URLFetch.PROPORTIONAL_ALGORITHM
    urlFetch.maxURLs = self.configVars[self.CONFIG_MAX_URLS_BATCH]
    urlFetchEvent = self.eventBuilder.build(DC_CONSTS.EVENT_TYPES.URL_FETCH, [urlFetch])
    # urlFetchEvent.cookie = ("sendURLFetchRequest")
    urlFetchEvent.cookie = {EventObjects.Batch.OPERATION_TYPE_NAME: EventObjects.Batch.TYPE_NORMAL_CRAWLER}
    # Send request URLFetch to SitesManager
    self.send(self.clientSitesManagerName, urlFetchEvent)
    logger.debug("The URLFetch request to SitesManager sent!")



  # #Send URLFetch(for incremental crawlering) request
  #
  #
  def sendIncrURLRequest(self):
    # add new state URL.STATUS_SELECTED_CRAWLING_INCREMENTAL and add it state for updating
    urlUpdate = EventObjects.URLUpdate(0, "", EventObjects.URLStatus.URL_TYPE_MD5, None,
                                       EventObjects.URL.STATUS_SELECTED_CRAWLING_INCREMENTAL)
    # Create URLFetch object
    conditionStr = ''' `Status` in (%s, %s, %s, %s) AND `Depth`<=%s \
    AND `State`=0 AND timediff(`TcDate`,`LastModified`)>=0 AND (NOW() - `TcDate`) > timediff(`TcDate`, `LastModified`)
    ''' % (EventObjects.URL.STATUS_CRAWLED, EventObjects.URL.STATUS_SELECTED_PROCESSING,
    EventObjects.URL.STATUS_PROCESSING, EventObjects.URL.STATUS_PROCESSED,
    self.configVars[DC_CONSTS.INCR_MAX_DEPTH_CONFIG_VAR_NAME])
    urlCriterions = {EventObjects.URLFetch.CRITERION_WHERE: conditionStr,
                     EventObjects.URLFetch.CRITERION_ORDER: "`Depth` ASC, `MRate` DESC, `UDate` DESC",
                     EventObjects.URLFetch.CRITERION_LIMIT:
                     str(self.configVars[DC_CONSTS.INCR_MAX_URLS_CONFIG_VAR_NAME])
                     }

    siteCriterions = {EventObjects.URLFetch.CRITERION_WHERE:
    "`ID` in (SELECT `Site_Id` FROM `sites_properties` WHERE `Name`='INCREMENTAL_CRAWLING' AND `Value`='1')"}

    urlFetch = EventObjects.URLFetch(None, urlCriterions, siteCriterions, urlUpdate)
    urlFetch.algorithm = EventObjects.URLFetch.PROPORTIONAL_ALGORITHM
    urlFetch.maxURLs = self.configVars[DC_CONSTS.INCR_MAX_URLS_CONFIG_VAR_NAME]
    urlFetchEvent = self.eventBuilder.build(DC_CONSTS.EVENT_TYPES.URL_FETCH, [urlFetch])
    urlFetchEvent.cookie = {"type": EventObjects.Batch.TYPE_INCR_CRAWLER}
    # Send request URLFetch to SitesManager
    self.send(self.clientSitesManagerName, urlFetchEvent)
    logger.debug("The URLFetch for incremental crawler request to SitesManager sent!")
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URL_FETCH_INCR_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)


  # #Process selected for crawling URLs return request
  #
  #
  def processSelectedURLsReturn(self):
    # Update to NEW state to return URLs in to the queue
    urlUpdate = EventObjects.URLUpdate(0, "", EventObjects.URLStatus.URL_TYPE_MD5, None,
                                       EventObjects.URL.STATUS_NEW)
    # Create URLFetch object with URLUpdate to update selected URLs state
    sitesCriterions = {EventObjects.URLFetch.CRITERION_WHERE: "`State`=" + str(EventObjects.Site.STATE_ACTIVE)}
    urlCriterions = {EventObjects.URLFetch.CRITERION_WHERE: "`Status` IN (" + \
                     str(EventObjects.URL.STATUS_SELECTED_CRAWLING) + "," + \
                     str(EventObjects.URL.STATUS_CRAWLING) + "," + \
                     str(EventObjects.URL.STATUS_SELECTED_PROCESSING) + "," + \
                     str(EventObjects.URL.STATUS_PROCESSING) + ") AND DATE_ADD(UDate, INTERVAL " + \
                     str(self.configVars[self.CONFIG_RET_URLS_TTL]) + " MINUTE) < NOW()",
                     EventObjects.URLFetch.CRITERION_ORDER: "`UDate` ASC",
                     EventObjects.URLFetch.CRITERION_LIMIT: str(self.configVars[self.CONFIG_RET_URLS_MAX_NUMBER])}
    urlFetch = EventObjects.URLFetch(None, urlCriterions, sitesCriterions, urlUpdate)
    urlFetch.algorithm = EventObjects.URLFetch.PROPORTIONAL_ALGORITHM
    urlFetch.maxURLs = self.configVars[self.CONFIG_RET_URLS_MAX_NUMBER]
    urlFetchEvent = self.eventBuilder.build(DC_CONSTS.EVENT_TYPES.URL_FETCH, [urlFetch])
    urlFetchEvent.cookie = {EventObjects.Batch.OPERATION_TYPE_NAME: EventObjects.Batch.TYPE_URLS_RETURN}
    # Send request URLFetch to SitesManager
    self.send(self.clientSitesManagerName, urlFetchEvent)
    logger.debug("The URLFetch request to SitesManager sent!")



  # #Get the DTM task state by involving one of two methods on DTM service
  #
  #@param taskId Id of DTM task
  def getDTMTaskState(self, taskId):
    taskState = None

    if self.configVars[self.CONFIG_BATCH_QUEUE_TASK_CHECK_METHOD] == self.DTM_TASK_CHECK_STATE_METHOD_STATUS:
      # Check the task status on DRCE EE hce-node
      logger.debug("Check state of taskId=" + str(taskId))
      checkTaskStateObj = dtm.EventObjects.CheckTaskState(taskId)
      checkStateEvent = self.eventBuilder.build(DTM_CONSTS.EVENT_TYPES.CHECK_TASK_STATE, checkTaskStateObj)
      eeResponseData = self.dtmdRequestExecute(checkStateEvent, self.configVars[self.CONFIG_DTMD_TIMEOUT])
      logger.debug("DTM CheckTaskState request finished, taskId=" + str(taskId))
      if eeResponseData is not None and type(eeResponseData) == type(dtm.EventObjects.EEResponseData(0)):
        taskState = eeResponseData.state
    else:
      # Get task status on DTM service
      logger.debug("Get status of taskId=" + str(taskId))
      getTasksStatusObj = dtm.EventObjects.GetTasksStatus([taskId])
      getTasksStatusEvent = self.eventBuilder.build(DTM_CONSTS.EVENT_TYPES.GET_TASK_STATUS, getTasksStatusObj)
      listTaskManagerFields = self.dtmdRequestExecute(getTasksStatusEvent, self.configVars[self.CONFIG_DTMD_TIMEOUT])
      logger.debug("DTM getTasksStatus request finished, taskId=" + str(taskId))
      if listTaskManagerFields is not None and isinstance(listTaskManagerFields, list) and\
        len(listTaskManagerFields) > 0:
        taskState = listTaskManagerFields[0].fields["state"]
      else:
        logger.error("DTM getTasksStatus taskId=" + str(taskId) + " returned wrong data:\n" + \
                     Utils.varDump(listTaskManagerFields))

    return taskState



  # #Process the DTM tasks queue
  #
  #
  def processDTMTasksQueue(self):
    tmpQueue = {}

    logger.debug("Batch tasks in queue:" + str(len(self.dtmTasksQueue)))
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_QUEUE_NAME, len(self.dtmTasksQueue),
                         self.STAT_FIELDS_OPERATION_SET)
    # Process the DTM tasks queue
    for taskId, taskBatch in self.dtmTasksQueue.items():
      batchState = self.getDTMTaskState(taskId)
      if batchState != None:
        logger.debug("Batch task state " + str(batchState) + ", taskId=" + str(taskId))
        if batchState == dtm.EventObjects.EEResponseData.TASK_STATE_FINISHED or\
           batchState == dtm.EventObjects.EEResponseData.TASK_STATE_CRASHED or\
           batchState == dtm.EventObjects.EEResponseData.TASK_STATE_TERMINATED or\
           batchState == dtm.EventObjects.EEResponseData.TASK_STATE_UNDEFINED or\
           batchState == dtm.EventObjects.EEResponseData.TASK_STATE_SET_ERROR or\
           batchState == dtm.EventObjects.EEResponseData.TASK_STATE_SCHEDULE_TRIES_EXCEEDED:
          # Delete task in DTM and task's data in EE (DRCE)
          deleteTaskObj = dtm.EventObjects.DeleteTask(taskId)
          deleteTaskEvent = self.eventBuilder.build(DTM_CONSTS.EVENT_TYPES.DELETE_TASK, deleteTaskObj)
          generalResponse = self.dtmdRequestExecute(deleteTaskEvent, self.configVars[self.CONFIG_DTMD_TIMEOUT])
          logger.debug("DTM DeleteTask request finished, taskId=" + str(taskId))
          if generalResponse is not None:
            if generalResponse.errorCode == dtm.EventObjects.GeneralResponse.ERROR_OK:
              logger.debug("DTM task deleted, taskId=" + str(taskId))
              if batchState == dtm.EventObjects.EEResponseData.TASK_STATE_FINISHED:
                logger.debug("batch:\n" + varDump(taskBatch) + "\n finished, taskId=" + str(taskId))
                self.processFinishedBatch(taskBatch)
              else:
                self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_FAULT_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
                self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URLS_FAULT_NAME, len(taskBatch.items),
                                     self.STAT_FIELDS_OPERATION_ADD)
                logger.debug("batch:" + varDump(taskBatch) + " not finished, state= " + str(batchState))
                # Send update URLs of finished batch on all nodes to get possibility to crawl them next time
                self.sendURLUpdate(taskBatch.items, taskBatch.id, False)
            else:
              # Save this batch to check it later
              tmpQueue[taskId] = taskBatch
              self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_DELETE_FAULT_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
              logger.error("DTM delete task taskId=" + str(taskId) + " error: " + str(generalResponse.errorCode) + \
                           " : " + generalResponse.errorMessage + ", statuses:" + varDump(generalResponse))
          else:
            # Save this batch to check it later
            tmpQueue[taskId] = taskBatch
            self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_DELETE_FAULT_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
            logger.error("DTM delete task error: wrong response or timeout, taskId=" + str(taskId) + "!")
        else:
          logger.debug("DTM task Id=" + str(taskId) + " state=" + str(batchState))
          if time.time() - taskBatch.QueuedTs > self.configVars[self.CONFIG_BATCH_QUEUE_TASK_TTL]:
            # Terminate task and delete it's data request
            deleteTaskObj = dtm.EventObjects.DeleteTask(taskId)
            deleteTaskObj.action = dtm.EventObjects.DeleteTask.ACTION_TERMINATE_TASK_AND_DELETE_DATA
            deleteTaskEvent = self.eventBuilder.build(DTM_CONSTS.EVENT_TYPES.DELETE_TASK, deleteTaskObj)
            generalResponse = self.dtmdRequestExecute(deleteTaskEvent, self.configVars[self.CONFIG_DTMD_TIMEOUT])
            logger.error("DTM task Id=" + str(taskId) + " terminated and removed from queue by TTL reason")
            # Send update URLs of finished batch on all nodes to get possibility to crawl them next time
            self.sendURLUpdate(taskBatch.items, taskBatch.id, False)
            self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_FAULT_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
            self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_FAULT_TTL_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
            self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_URLS_FAULT_NAME, len(taskBatch.items),
                                 self.STAT_FIELDS_OPERATION_ADD)
          else:
            # Save this batch to check it later
            tmpQueue[taskId] = taskBatch
            logger.debug("DTM task Id=" + str(taskId) + " still in queue")
      else:
        # Save this batch to check it later
        tmpQueue[taskId] = taskBatch
        self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_CHECK_FAULT_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
        logger.error("DTM check task state error: wrong response or timeout, taskId=" + str(taskId) + "!")

    self.dtmTasksQueue = tmpQueue
    self.updateStatField(DC_CONSTS.BATCHES_CRAWL_COUNTER_QUEUE_NAME, len(self.dtmTasksQueue),
                         self.STAT_FIELDS_OPERATION_SET)
    logger.debug("The DTM tasks queue processing finished, batch tasks in queue " + str(len(self.dtmTasksQueue)))



  # #Do some post batch processing after batch was successfully finished
  #
  # @param taskBatch the Batch object
  #
  def processFinishedBatch(self, taskBatch):
    if self.configVars[self.CONFIG_CRAWLED_URLS_STRATEGY] == 0:
      self.sendURLUpdate(taskBatch.items, taskBatch.id, True)
      logger.debug("Send update URLs from batch for all foreign hosts by the Batch_Id")
    else:
      self.sendURLDelete(taskBatch.items, taskBatch.id)
      logger.debug("Send delete URLs from batch for all foreign hosts by the Batch_Id")



  # #Execute DTMD task request
  #
  # @param requestEvent The request event to send to DTMD
  # @param timeout The DTMD request timeout
  #
  def dtmdRequestExecute(self, requestEvent, timeout, maxTries=100):
    ret = None
    if maxTries < 0:
      maxTries = 0

    try:
      # Send DTMD request
      self.dtmdConnection.send(requestEvent)

      for i in range(maxTries + 1):
        # Poll DTMD connection
        if self.dtmdConnection.poll(int(timeout)) == 0:
          logger.error("DTMD request timeout reached " + str(timeout) + "!")
          break
        else:
          # Recv DTMD response
          retEvent = self.dtmdConnection.recv()
          if retEvent != None:
            # Get response object
            if type(retEvent.eventObj) == type(dtm.EventObjects.EEResponseData(0)) or\
              type(retEvent.eventObj) == type(dtm.EventObjects.GeneralResponse()) or\
              isinstance(retEvent.eventObj, list):
              if retEvent.uid == requestEvent.uid:
                ret = retEvent.eventObj
                break
              else:
                logger.error("DTMD returned wrong object uid: " + str(retEvent.uid) + " but " + \
                             str(requestEvent.uid) + " expected, iteration " + str(i) + "!")
            else:
              logger.error("DTMD returned wrong object type: " + str(type(retEvent.eventObj)) + "!")
          else:
            logger.error("DTMD returned None event!")
    except Exception, e:
      logger.error("DTMD request execution exception: " + e.message + "!")

    logger.debug("The DTMD request finished!")

    return ret



  # #Send URL update for batch URLs
  #
  # @param batchItemsList List of BatchItem objects
  # @param batchId the Batch Id
  # @param batchState state of batch operation False - means return URLs to New state, True - means set crawled state
  def sendURLUpdate(self, batchItemsList, batchId, batchState):
    urlsList = []

    # Prepare URLs list to update
    for batchItem in batchItemsList:
      # Set state value depends on update reason - crawled successfully or not
      if batchState is True:
        status = EventObjects.URL.STATUS_CRAWLED
        sqlExpression = SQLExpression("`URLMd5`='" + str(batchItem.urlId) + "' AND `Batch_Id` <>" + str(batchId) + \
                                      " AND `Status` NOT IN (" + str(EventObjects.URL.STATUS_CRAWLED) + \
                                      "," + str(EventObjects.URL.STATUS_SELECTED_PROCESSING) + ")")
      else:
        status = EventObjects.URL.STATUS_NEW
        sqlExpression = SQLExpression("`URLMd5`='" + str(batchItem.urlId) + "' AND `Status` IN (" +
                                      str(EventObjects.URL.STATUS_SELECTED_CRAWLING) + "," + \
                                      str(EventObjects.URL.STATUS_SELECTED_CRAWLING_INCREMENTAL) + ")")

      urlUpdate = EventObjects.URLUpdate(batchItem.siteId, batchItem.urlId, EventObjects.URLStatus.URL_TYPE_MD5,
                                         None, status)
      urlUpdate.criterions[EventObjects.URLFetch.CRITERION_WHERE] = sqlExpression
      logger.debug("URLUpdate: " + varDump(urlUpdate))
      urlsList.append(urlUpdate)

    # Make URLUpdate event
    urlUpdateEvent = self.eventBuilder.build(DC_CONSTS.EVENT_TYPES.URL_UPDATE, urlsList)
    # Send request URLUpdate to SitesManager
    self.send(self.clientSitesManagerName, urlUpdateEvent)
    logger.debug("The URLUpdate request to SitesManager sent!")



  # #Send URL delete for batch URLs that is not crawled and stays in SELECTED_FOR_CRAWLING state (2)
  #
  # @param batchItemsList List of BatchItem objects
  # @param batchId Id of the batch
  def sendURLDelete(self, batchItemsList, batchId):
    urlsList = []

    # Prepare URLs list to delete
    for batchItem in batchItemsList:
      sqlExpression = SQLExpression("`ParentMd5`<>'' AND `URLMd5`='" + str(batchItem.urlId) + "' AND `Batch_Id`<>" + \
                                     str(batchId))
      urlDelete = EventObjects.URLDelete(batchItem.siteId, None, EventObjects.URLStatus.URL_TYPE_URL,
                                         {EventObjects.URLFetch.CRITERION_WHERE:sqlExpression,
                                          EventObjects.URLFetch.CRITERION_LIMIT:1})
      logger.debug("URLDelete: " + varDump(urlDelete))
      urlsList.append(urlDelete)

    # Make URLDelete event
    urlDeleteEvent = self.eventBuilder.build(DC_CONSTS.EVENT_TYPES.URL_DELETE, urlsList)
    # Send request URLDelete to SitesManager
    self.send(self.clientSitesManagerName, urlDeleteEvent)
    logger.debug("The URLDelete request to SitesManager sent!")



  # #onURLDeleteResponse event handler
  #
  # @param event instance of Event object
  def onURLDeleteResponse(self, event):
    try:
      logger.debug("Reply received on URL delete.")
      clientResponse = event.eventObj
      if clientResponse.errorCode == EventObjects.ClientResponse.STATUS_OK:
        if len(clientResponse.itemsList) > 0:
          for clientResponseItem in clientResponse.itemsList:
            if clientResponseItem.errorCode != EventObjects.ClientResponseItem.STATUS_OK:
              logger.error("URLDelete response error: " + str(clientResponseItem.errorCode) + " : " + \
                           clientResponseItem.errorMessage + ", host:" + clientResponseItem.host + ", port:" + \
                           clientResponseItem.port + ", node:" + clientResponseItem.node + "!")
        else:
          logger.error("URLDelete response empty list!")
      else:
        logger.error("URLDelete response error:" + str(clientResponse.errorCode) + " : " + clientResponse.errorMessage)
    except Exception as err:
      logger.error("Exception: " + str(err.message) + "\n" + Utils.getTracebackInfo())

