'''
HCE project, Python bindings, Distributed Crawler application.
SitesManager object and related classes definitions.

@package: dc
@author bgv bgv.hce@gmail.com
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
'''

import time
import ctypes
import logging
import zlib

# import pickle
try:
  import cPickle as pickle
except ImportError:
  import pickle

from app.BaseServerManager import BaseServerManager
from dc import EventObjects
from dc.Constants import EVENT_TYPES
from drce.CommandConvertor import CommandConvertor
from drce.Commands import Session
from drce.Commands import TaskExecuteRequest
from drce.Commands import TaskExecuteStruct
from drce.DRCEManager import ConnectionTimeout, TransportInternalErr, CommandExecutorErr
from drce.DRCEManager import DRCEManager
from drce.DRCEManager import HostParams
from transport.ConnectionBuilderLight import ConnectionBuilderLight
from transport.IDGenerator import IDGenerator
from transport.UIDGenerator import UIDGenerator
import dc.Constants as DC_CONSTS
import transport.Consts as TRANSPORT_CONSTS
import app.Utils as Utils  # pylint: disable=F0401


# Logger initialization
logger = logging.getLogger(DC_CONSTS.LOGGER_NAME)


# #The SitesManager class, is a main interface of DC application and distributed sites database.
#
# This object is a main interface of DC application and DRCE cluster that operates with distributed DB units
# It supports complete DRCE protocol requests and process responses from DRCE cluster.
#
class SitesManager(BaseServerManager):

  DRCE_REDUCER_TTL = 3000000
  SITE_PROPERTIES_RECRAWL_WHERE_NAME = "RECRAWL_WHERE"
  SITE_PROPERTIES_RECRAWL_DELETE_WHERE_NAME = "RECRAWL_DELETE_WHERE"
  SITE_PROPERTIES_RECRAWL_DELETE_NAME = "RECRAWL_DELETE"

  # Configuration settings options names
  CONFIG_SERVER = "server"
  CONFIG_DRCE_HOST = "DRCEHost"
  CONFIG_DRCE_PORT = "DRCEPort"
  CONFIG_DRCE_TIMEOUT = "DRCETimeout"
  CONFIG_DRCE_DB_APP_NAME = "DRCEDBAppName"
  CONFIG_RECRAWL_SITES_MAX = "RecrawlSitesMax"
  CONFIG_RECRAWL_SITES_PERIOD = "RecrawlSitesPeriod"
  CONFIG_POLLING_TIMEOUT = "PollingTimeout"
  CONFIG_DEFAULT_RECRAWL_UPDATE_CRITERION = "DefaultRecrawUpdatelCriterion"
  CONFIG_DEFAULT_RECRAWL_DELETE_OLD = "DefaultRecrawDeleteOld"
  CONFIG_DEFAULT_RECRAWL_DELETE_OLD_CRITERION = "DefaultRecrawDeleteOldCriterion"
  CONFIG_RESOURCE_AGING_PERIOD = "ResourcesAgingPeriod"
  CONFIG_DRCE_ROUTE = "DRCERoute"

  # #constructor
  # initialize fields
  #
  # @param configParser config parser object
  # @param connectBuilderLight connection builder light
  #
  def __init__(self, configParser, connectionBuilderLight=None):
    super(SitesManager, self).__init__()

    # Sites re-crawl counter name for stat variables
    self.updateStatField(DC_CONSTS.SITES_RECRAWL_COUNTER_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Sites re-crawl sites updated counter name for stat variables
    self.updateStatField(DC_CONSTS.SITES_RECRAWL_UPDATED_COUNTER_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Sites re-crawl sites deleted counter name for stat variables
    self.updateStatField(DC_CONSTS.SITES_RECRAWL_DELETED_COUNTER_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)
    # Sites DRCE requests counter name for stat variables
    self.updateStatField(DC_CONSTS.SITES_DRCE_COUNTER_NAME, 0, self.STAT_FIELDS_OPERATION_INIT)

    # Instantiate the connection builder light if not set
    if connectionBuilderLight == None:
      connectionBuilderLight = ConnectionBuilderLight()

    className = self.__class__.__name__

    # Get configuration settings
    self.serverName = configParser.get(className, self.CONFIG_SERVER)
    self.DRCEDBAppName = configParser.get(className, self.CONFIG_DRCE_DB_APP_NAME)

    # Create connections and raise bind or connect actions for correspondent connection type
    serverConnection = connectionBuilderLight.build(TRANSPORT_CONSTS.SERVER_CONNECT, self.serverName)

    # Add connections to the polling set
    self.addConnection(self.serverName, serverConnection)

    # Set handlers for all events of SITE and URL operations
    self.eventTypes = {EVENT_TYPES.SITE_NEW:EVENT_TYPES.SITE_NEW_RESPONSE,
                       EVENT_TYPES.SITE_UPDATE:EVENT_TYPES.SITE_UPDATE_RESPONSE,
                       EVENT_TYPES.SITE_STATUS:EVENT_TYPES.SITE_STATUS_RESPONSE,
                       EVENT_TYPES.SITE_DELETE:EVENT_TYPES.SITE_DELETE_RESPONSE,
                       EVENT_TYPES.SITE_CLEANUP:EVENT_TYPES.SITE_CLEANUP_RESPONSE,
                       EVENT_TYPES.SITE_FIND:EVENT_TYPES.SITE_FIND_RESPONSE,
                       EVENT_TYPES.URL_NEW:EVENT_TYPES.URL_NEW_RESPONSE,
                       EVENT_TYPES.URL_STATUS:EVENT_TYPES.URL_STATUS_RESPONSE,
                       EVENT_TYPES.URL_UPDATE:EVENT_TYPES.URL_UPDATE_RESPONSE,
                       EVENT_TYPES.URL_DELETE:EVENT_TYPES.URL_DELETE_RESPONSE,
                       EVENT_TYPES.URL_FETCH:EVENT_TYPES.URL_FETCH_RESPONSE,
                       EVENT_TYPES.URL_CLEANUP:EVENT_TYPES.URL_CLEANUP_RESPONSE,
                       EVENT_TYPES.URL_CONTENT:EVENT_TYPES.URL_CONTENT_RESPONSE}
    for ret in self.eventTypes:
      # Set event handler for SITE_NEW event
      self.setEventHandler(ret, self.onEventsHandler)

    # Initialize DRCE API
    self.configVars[self.CONFIG_DRCE_TIMEOUT] = configParser.getint(className, self.CONFIG_DRCE_TIMEOUT)
    self.configVars[self.CONFIG_DRCE_ROUTE] = configParser.get(className, self.CONFIG_DRCE_ROUTE)
    self.drceHost = configParser.get(className, self.CONFIG_DRCE_HOST)
    self.drcePort = configParser.get(className, self.CONFIG_DRCE_PORT)
    hostParams = HostParams(self.drceHost, self.drcePort)
    self.drceManager = DRCEManager()
    self.drceManager.activate_host(hostParams)
    self.drceIdGenerator = UIDGenerator()
    self.drceCommandConvertor = CommandConvertor()

    #Init config vars storage for auto re-crawl
    self.configVars[self.CONFIG_RECRAWL_SITES_MAX] = configParser.getint(className, self.CONFIG_RECRAWL_SITES_MAX)
    self.configVars[self.CONFIG_RECRAWL_SITES_PERIOD] = configParser.getint(className, self.CONFIG_RECRAWL_SITES_PERIOD)
    self.processRecrawlLastTs = time.time()

    # Set connections poll timeout, defines period of HCE cluster monitoring cycle, msec
    self.configVars[self.POLL_TIMEOUT_CONFIG_VAR_NAME] = configParser.getint(className, self.CONFIG_POLLING_TIMEOUT)

    # Init default re-crawl update criterion
    self.configVars[self.CONFIG_DEFAULT_RECRAWL_UPDATE_CRITERION] = \
         configParser.get(className, self.CONFIG_DEFAULT_RECRAWL_UPDATE_CRITERION)

    # Init default re-crawl delete old URLs operation
    self.configVars[self.CONFIG_DEFAULT_RECRAWL_DELETE_OLD] = \
         configParser.getint(className, self.CONFIG_DEFAULT_RECRAWL_DELETE_OLD)
    # Init default re-crawl delete old URLs operation's criterion
    self.configVars[self.CONFIG_DEFAULT_RECRAWL_DELETE_OLD_CRITERION] = \
         configParser.get(className, self.CONFIG_DEFAULT_RECRAWL_DELETE_OLD_CRITERION)

    # Init default resource aging settings
    self.configVars[self.CONFIG_RESOURCE_AGING_PERIOD] = configParser.getint(className,
                                                                             self.CONFIG_RESOURCE_AGING_PERIOD)
    self.processResourcesAgingLastTs = time.time()


  # #Events wait timeout handler, for timeout state of the connections polling. Executes periodical check of DTM tasks
  # and initiate the main crawling iteration cycle
  #
  def on_poll_timeout(self):
    logger.debug("Periodic iteration started.")
    try:
      #Process periodic re-crawling
      if time.time() - self.processRecrawlLastTs > self.configVars[self.CONFIG_RECRAWL_SITES_PERIOD]:
        logger.debug("Now time to perform re-crawl, interval " + str(self.configVars[self.CONFIG_RECRAWL_SITES_PERIOD]))
        self.processRecrawlLastTs = time.time()
        self.processRecrawling()

      #Process resources delete by aging criterion with TTL defined in sites_properties[RESOURCE_TTL]
      if time.time() - self.processResourcesAgingLastTs > self.configVars[self.CONFIG_RESOURCE_AGING_PERIOD]:
        logger.debug("Now time to perform resources aging, interval " + \
                     str(self.configVars[self.CONFIG_RESOURCE_AGING_PERIOD]))
        self.processResourcesAgingLastTs = time.time()
        self.processResourcesAging()

    except Exception as err:
      logger.error("Exception: " + str(err.message) + "\n" + Utils.getTracebackInfo())

    logger.debug("Periodic iteration finished.")



  # #onEventsHandler event handler
  #
  # @param event instance of Event object
  def onEventsHandler(self, event):
    # Process send request
    clientResponseObj = self.processDRCERequest(self.prepareDRCERequest(event.eventType, event.eventObj))
    logger.info("ClientResponseObj object:\n" + Utils.varDump(clientResponseObj))
    # Prepare reply event
    replyEvent = self.eventBuilder.build(self.eventTypes[event.eventType], clientResponseObj)
    # Send reply
    self.reply(event, replyEvent)
    logger.info("Reply sent")



  # #Prepare DRCE request
  #
  # @param eventType event type from Constants
  # @param eventObj instance of Event object
  # @return the TaskExecuteStruct object instance
  def prepareDRCERequest(self, eventType, eventObj):
    # Prepare DRCE request data
    drceSyncTasksCoverObj = DC_CONSTS.DRCESyncTasksCover(eventType, eventObj)
    # Prepare DRCE request object
    taskExecuteStruct = TaskExecuteStruct()
    taskExecuteStruct.command = self.DRCEDBAppName
    taskExecuteStruct.input = pickle.dumps(drceSyncTasksCoverObj)
    taskExecuteStruct.session = Session(Session.TMODE_SYNC)
    logger.debug("DRCE taskExecuteStruct:\n" + Utils.varDump(taskExecuteStruct))

    return taskExecuteStruct



  # #Request action processor for DRCE DB cluster
  #
  # @param taskExecuteStruct object
  def processDRCERequest(self, taskExecuteStruct):
    # Create DRCE TaskExecuteRequest object
    idGenerator = IDGenerator()
    taskId = ctypes.c_uint32(zlib.crc32(idGenerator.get_connection_uid())).value
    taskExecuteRequest = TaskExecuteRequest(taskId)
    if self.configVars[self.CONFIG_DRCE_ROUTE] != "":
      taskExecuteRequest.route = self.configVars[self.CONFIG_DRCE_ROUTE]
    # Set taskExecuteRequest fields
    taskExecuteRequest.data = taskExecuteStruct
    logger.info("Sending sync task id:" + str(taskId) + " to DRCE router!")
    # Send request to DRCE Cluster router
    response = self.sendToDRCERouter(taskExecuteRequest)
    logger.info("Received response on sync task from DRCE router!")
    logger.debug("Response: %s", Utils.varDump(response))

    # Create new client response object
    clientResponse = EventObjects.ClientResponse()

    # Check response returned
    if response == None:
      clientResponse.errorCode = EventObjects.ClientResponse.STATUS_ERROR_NONE
      clientResponse.errorMessage = "Response error, None returned from DRCE, possible timeout " + \
                                    str(self.configVars[self.CONFIG_DRCE_TIMEOUT]) + " msec!"
      logger.error(clientResponse.errorMessage)
    else:
      if len(response.items) == 0:
        clientResponse.errorCode = EventObjects.ClientResponse.STATUS_ERROR_EMPTY_LIST
        clientResponse.errorMessage = "Response error, empty list returned from DRCE, possible no one node in cluster!"
        logger.error(clientResponse.errorMessage)
      else:
        for item in response.items:
          # New ClientResponseItem object
          clientResponseItem = EventObjects.ClientResponseItem(None)
          # If some error in response item or cli application exit status
          if item.error_code > 0 or item.exit_status > 0:
            clientResponseItem.errorCode = clientResponseItem.STATUS_ERROR_DRCE
            clientResponseItem.errorMessage = "Response item error error_message=" + item.error_message + \
                                              ", error_code=" + str(item.error_code) + \
                                              ", exit_status=" + str(item.exit_status) + \
                                              ", stderror=" + str(item.stderror)
            logger.error(clientResponseItem.errorMessage)
          else:
            # Try to restore serialized response object from dump
            try:
              drceSyncTasksCover = pickle.loads(item.stdout)
              clientResponseItem.itemObject = drceSyncTasksCover.eventObject
            except Exception as e:
              clientResponseItem.errorCode = EventObjects.ClientResponseItem.STATUS_ERROR_RESTORE_OBJECT
              clientResponseItem.errorMessage = EventObjects.ClientResponseItem.MSG_ERROR_RESTORE_OBJECT + "\n" + \
                                                str(e.message) + "\nstdout=" + str(item.stdout) + \
                                                ", stderror=" + str(item.stderror)
              logger.error(clientResponseItem.errorMessage)
          # Set all information fields in any case
          clientResponseItem.id = item.id
          clientResponseItem.host = item.host
          clientResponseItem.port = item.port
          clientResponseItem.node = item.node
          clientResponseItem.time = item.time
          # Add ClientResponseItem object
          clientResponse.itemsList.append(clientResponseItem)

    return clientResponse



  # #Send to send to DRCE Router transport router connection
  #
  # @param messageBody body of the message
  # @return EEResponseData object instance
  def sendToDRCERouter(self, request):
    self.updateStatField(DC_CONSTS.SITES_DRCE_COUNTER_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
    logger.info("DRCE router sending with timeout=" + str(self.configVars[self.CONFIG_DRCE_TIMEOUT]) + \
                ", host:" + str(self.drceHost) + ", port:" + str(self.drcePort))
    # Try to execute request
    try:
      response = self.drceManager.process(request, self.configVars[self.CONFIG_DRCE_TIMEOUT], self.DRCE_REDUCER_TTL)
    except (ConnectionTimeout, TransportInternalErr, CommandExecutorErr) as err:
      response = None
      logger.error("DRCE router send error : " + str(err.message))

    logger.info("DRCE router sent!")

    return response



  # #Process periodic auto re-crawling
  #
  #
  def processRecrawling(self):
    self.updateStatField(DC_CONSTS.SITES_RECRAWL_COUNTER_NAME, 1, self.STAT_FIELDS_OPERATION_ADD)
    #Find sites that need to be re-crawled
    crit = {EventObjects.SiteFind.CRITERION_WHERE: "`State`=1 AND `RecrawlDate` IS NOT NULL AND NOW()>`RecrawlDate`",
            EventObjects.SiteFind.CRITERION_ORDER: "`RecrawlDate` ASC",
            EventObjects.SiteFind.CRITERION_LIMIT: str(self.configVars[self.CONFIG_RECRAWL_SITES_MAX])}
    siteFind = EventObjects.SiteFind(None, crit)
    logger.debug("Send DRCE request SITE_FIND")
    clientResponse = self.processDRCERequest(self.prepareDRCERequest(EVENT_TYPES.SITE_FIND, siteFind))
    logger.debug("clientResponse:" + Utils.varDump(clientResponse))
    #Process results and create unique sites dict
    sites = self.getSitesFromClientResponseItems(clientResponse.itemsList)
    urlUpdateList = []
    urlDeleteList = []
    #TODO: Update per site can be improved if multi-site update will be supported
    for siteId in sites.keys():
      logger.debug("Site[" + siteId + "]:\n" + Utils.varDump(sites[siteId]))
      siteUpdate = EventObjects.SiteUpdate(siteId)
      siteUpdate.iterations = EventObjects.SQLExpression("Iterations+1")
      siteUpdate.tcDate = EventObjects.SQLExpression("Now()")
      siteUpdate.uDate = siteUpdate.tcDate
      if sites[siteId].recrawlPeriod > 0:
        #Set next re-crawl date if periodic re-crawl
        #siteUpdate.recrawlDate = EventObjects.SQLExpression("DATE_ADD(Now(), INTERVAL `RecrawlPeriod` MINUTE)")
        siteUpdate.recrawlDate = EventObjects.SQLExpression("from_unixtime(UNIX_TIMESTAMP()+`RecrawlPeriod`*60)")
      else:
        #Disable re-crawl
        siteUpdate.recrawlDate = EventObjects.SQLExpression("NULL")
      #Update site
      clientResponse = self.processDRCERequest(self.prepareDRCERequest(EVENT_TYPES.SITE_UPDATE, siteUpdate))
      #Check responses on errors
      self.logGeneralResponseResults(clientResponse)
      urlUpdate = EventObjects.URLUpdate(siteId, "")
      urlUpdate.url = None
      urlUpdate.urlMd5 = None
      urlUpdate.status = EventObjects.URLUpdate.STATUS_NEW
      crit = EventObjects.Site.getFromProperties(sites[siteId].properties, self.SITE_PROPERTIES_RECRAWL_WHERE_NAME)
      if crit is None or crit == "":
        crit = self.configVars[self.CONFIG_DEFAULT_RECRAWL_UPDATE_CRITERION]
        logger.debug("Default update criterion: " + str(crit))
      else:
        logger.debug("Custom site update criterion: " + str(crit))
      urlUpdate.criterions = {EventObjects.URLFetch.CRITERION_WHERE : crit}
      urlUpdateList.append(urlUpdate)

      #Delete old URLs from prev. re-crawling iteration
      if self.configVars[self.CONFIG_DEFAULT_RECRAWL_DELETE_OLD] > 0:
        sqlExpression = EventObjects.SQLExpression(self.configVars[self.CONFIG_DEFAULT_RECRAWL_DELETE_OLD_CRITERION])
        siteDelOld = EventObjects.Site.getFromProperties(sites[siteId].properties,
                                                         self.SITE_PROPERTIES_RECRAWL_DELETE_NAME)
        if siteDelOld is None or siteDelOld != "0":
          siteSqlExpression = EventObjects.Site.getFromProperties(sites[siteId].properties,
                                                    self.SITE_PROPERTIES_RECRAWL_DELETE_WHERE_NAME)
          logger.debug("Site expression: " + str(siteSqlExpression))
          if siteSqlExpression is not None and siteSqlExpression != "":
            sqlExpression = siteSqlExpression
            logger.debug("Custom expression set: " + str(sqlExpression))
        else:
          logger.debug("Site delete old: " + str(siteDelOld))

        urlDelete = EventObjects.URLDelete(siteId, None, EventObjects.URLStatus.URL_TYPE_MD5,
                                           {EventObjects.URLFetch.CRITERION_WHERE:sqlExpression})
        logger.debug("Old URLs delete due re-crawl: " + Utils.varDump(urlDelete))
        urlDeleteList.append(urlDelete)

    if self.configVars[self.CONFIG_DEFAULT_RECRAWL_DELETE_OLD] > 0:
      #Delete old URLs from prev. re-crawling
      clientResponse = self.processDRCERequest(self.prepareDRCERequest(EVENT_TYPES.URL_DELETE, urlDeleteList))
      self.updateStatField(DC_CONSTS.SITES_RECRAWL_DELETED_COUNTER_NAME, len(urlDeleteList),
                           self.STAT_FIELDS_OPERATION_ADD)
      #Check responses on errors
      self.logGeneralResponseResults(clientResponse)

    #Update URLs to push them to start crawling
    clientResponse = self.processDRCERequest(self.prepareDRCERequest(EVENT_TYPES.URL_UPDATE, urlUpdateList))
    self.updateStatField(DC_CONSTS.SITES_RECRAWL_UPDATED_COUNTER_NAME, len(urlUpdateList),
                         self.STAT_FIELDS_OPERATION_ADD)
    #Check responses on errors
    self.logGeneralResponseResults(clientResponse)



  ##Get sites dict {SiteId:Site} from the ClientResponse object items
  #
  # @param clientResponseItems The list of ClientResponseItem objects
  def getSitesFromClientResponseItems(self, clientResponseItems):
    batchItemsCounter = 0
    batchItemsTotalCounter = 0
    uniqueSitesDic = {}

    for item in clientResponseItems:
      if item.errorCode == EventObjects.ClientResponseItem.STATUS_OK:
        if isinstance(item.itemObject, list):
          for site in item.itemObject:
            batchItemsTotalCounter = batchItemsTotalCounter + 1
            if isinstance(site, EventObjects.Site):
              if str(site.id) not in uniqueSitesDic:
                uniqueSitesDic[str(site.id)] = site
                batchItemsCounter = batchItemsCounter + 1
            else:
              logger.error("Wrong object type in the itemObject.item: " + str(type(site)) + \
                           " but 'Site' expected")
        else:
          logger.error("Wrong object type in the ClientResponseItem.itemObject: " + str(type(item.itemObject)) + \
                       " but 'list' expected")
      else:
        logger.debug("ClientResponseItem error: " + str(item.errorCode) + " : " + item.errorMessage)

    logger.debug("Unique sites: " + str(batchItemsCounter) + ", total sites: " + str(batchItemsTotalCounter))

    return uniqueSitesDic



  ##Check the ClientResponse object with GeneralResponse objects in items list, log errors
  #
  # @param ClientResponse object
  def logGeneralResponseResults(self, clientResponse):
    if isinstance(clientResponse, EventObjects.ClientResponse):
      if clientResponse.errorCode > 0:
        logger.error("clientResponse.errorCode:" + str(clientResponse.errorCode) + ":" + clientResponse.errorMessage)
      for clientResponseItem in clientResponse.itemsList:
        if isinstance(clientResponseItem, EventObjects.ClientResponseItem):
          if clientResponseItem.errorCode != EventObjects.ClientResponseItem.STATUS_OK:
            logger.error("ClientResponseItem error: " + str(clientResponseItem.errorCode) + " : " + \
                         clientResponseItem.errorMessage + "\n" + Utils.varDump(clientResponseItem))
        else:
          logger.error("Wrong type: " + str(type(clientResponseItem)) + ", expected ClientResponseItem\n" + \
                       Utils.varDump(clientResponseItem))
    else:
      logger.error("Wrong type: " + str(type(clientResponse)) + ", expected ClientResponse\n" + \
                   Utils.varDump(clientResponse))



  # #Process periodic resources delete as aging with TTL from sites_properties[RESOURCE_TTL]
  #
  #
  def processResourcesAging(self):
    pass

