/**
 * @file ClientsQueueManager.hpp
 *
 * HCE project node clients of PPP manager.
 * Implements functionality for zmq server connection of registered clients management.
 * Manage clients registration and life time by client name. Used in all kind of PPP clients registration to identify clients, broadcast
 * messages send, heartbit messages send and so on.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_CLIENTS_QUEUE_MANAGER_INCLUDED__
#define __HCE_CLIENTS_QUEUE_MANAGER_INCLUDED__

#include "HandlerTypes.hpp"

namespace HCE {
  namespace handlers {

    class ClientsQueueManager {
      public:
        //Purge mode: 0 - purge client from queue if expired, 1 - not purge but increment counter only
        ClientsQueueManager(std::vector<ClientWorkerItem>* queue, unsigned int purgeMode) :
            _refreshNotFoundCounter(0), _purgeMode(purgeMode), _purgeCounter(0){
          _queue = queue;
        }

        //!<Insert client at end of queue, reset expiry, client must not already be in queue
        void append(const std::string &identity, bool checkPresence = true){
          bool found = false;

          if(checkPresence){
            //Check allworkers by identity
            for(std::vector<ClientWorkerItem>::iterator it = _queue->begin(); it < _queue->end(); ++it){
              if(it->identity.compare(identity) == 0){
                found = true;
                break;
              }
            }
          }

          //Not found
          if(!found){
            //Add new worker
            ClientWorkerItem worker = {identity, (s_clock() + HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS), 0, 0};
            _queue->push_back(worker);

            //Reset counters to balance requests proportional way after new worker insertion
            resetFrequencies();
          }
        }

        //!<Remove client from queue, if present
        void remove(const std::string &identity){
          for(std::vector<ClientWorkerItem>::iterator it = _queue->begin(); it < _queue->end(); ++it){
            if(it->identity.compare(identity) == 0){
              it = _queue->erase(it);
              break;
            }
          }
        }

        //!<Reset worker expiry, worker must be present
        bool refresh(const std::string &identity){
          bool found = false;

          for(std::vector<ClientWorkerItem>::iterator it = _queue->begin(); it < _queue->end(); ++it){
            if(it->identity.compare(identity) == 0){
              it->expiry = s_clock() + HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS;
              found = true;
              break;
            }
          }

          //TODO: experimental
          if(!found){
            //If not found - add as new and fresh
            append(identity);

            _refreshNotFoundCounter++;
          }

          return found;
        }

        //!<Look for & kill expired clients
        std::string purge(void){
          std::stringstream ret("");

          int64_t clock = s_clock();
          for(std::vector<ClientWorkerItem>::iterator it = _queue->begin(); it < _queue->end(); ++it){
           if(clock > it->expiry){
             ret << "{\"identity\":\"" << it->identity << "\",\"expiry\":" << it->expiry << ",\"count\":" << it->count << ",\"frequency\":" << it->frequency << "},";
             if(_purgeMode == 0){
               it = _queue->erase(it) - 1;
             }
             _purgeCounter++;
           }
          }
          return ret.str();
        }

        //!<Existance registered client check
        bool exists(const std::string &identity){
          bool ret = false;

          for(std::vector<ClientWorkerItem>::iterator it = _queue->begin(); it < _queue->end(); ++it){
            if(it->identity.compare(identity) == 0){
              ret = true;
              break;
            }
          }

          return ret;
        }

        //!<Returns dump of registered clients and current stats data
        std::string getDump() const{
          std::stringstream stream;

          stream << _purgeCounter << "," << _refreshNotFoundCounter << "," << _queue->size() << ",";
          for(std::vector<ClientWorkerItem>::iterator it = _queue->begin(); it < _queue->end(); ++it){
            stream << it->identity << " " << it->count << " " << it->frequency << " " << it->expiry << ",";
          }

          return stream.str();
        }

        //!<Resets requests frequencies, increment counters
        void resetFrequencies(void){
          for(std::vector<ClientWorkerItem>::iterator it = _queue->begin(); it < _queue->end(); ++it){
            it->count += it->frequency;
            it->frequency = 0;
          }
        }

        //!<Get purge mode
        unsigned int getPurgeMode(void) const{
          return _purgeMode;
        }

        //!<Set purge mode
        void setPurgeMode(const unsigned int purgeMode){
          _purgeMode = purgeMode;
        }

      private:
        //!<Client workers
        std::vector<ClientWorkerItem>* _queue;
        //!<Stat counter of cases "client not found" in refresh action
        unsigned long long _refreshNotFoundCounter;
        unsigned int _purgeMode;
        unsigned long long _purgeCounter;
    };
  }
}

#endif /* __HCE_CLIENTS_QUEUE_MANAGER_INCLUDED__ */
