#include "DataClientProxy.hpp"

namespace HCE {
  namespace handlers {

    DataClientProxy::DataClientProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString, int64_t startedAt, unsigned char logPriority) :
        Handler(name, ctx, connectionString, startedAt, logPriority), _clientIdentity(clientIdentity){

      //!<Init all the rest
      initialize();
    }

    DataClientProxy::~DataClientProxy(){
      deinitialize();
    }

    int DataClientProxy::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << _clientIdentity << "] with connection [" << _connectionString << "]" << flush;

      //!<Init sockets
      _inprocProxySock = NULL;
      _inprocReducerOutSock = NULL;
      _dataClientSock = NULL;
      _dataClientSockConnected = false;

      //!<Heartbeat at time mark
      _heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;
      //!<Heartbeated time mark
      _heartbeatedAt = s_clock();

      //!<Stat vars
      _statRequests = 0;
      _statResponses = 0;

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildClientConnection(_inprocProxySock, NODE_CS_INPROC_PROXY, ZMQ_DEALER);

      //!<Rebuild inproc_reducer_out connection
      rebuildClientConnection(_inprocReducerOutSock, NODE_CS_INPROC_REDUCER_OUT, ZMQ_DEALER);

      //!<Create dataclient socket, connect to upper node and send READY message
      rebuildDataClientSocket();

      return true;
    }

    void DataClientProxy::deinitialize(void){
      if(_inprocProxySock){
        if(_inprocProxySock->connected()){
          _inprocProxySock->close();
        }
        delete _inprocProxySock;
        _inprocProxySock = NULL;
      }

      if(_inprocReducerOutSock){
        if(_inprocReducerOutSock->connected()){
          _inprocReducerOutSock->close();
        }
        delete _inprocReducerOutSock;
        _inprocReducerOutSock = NULL;
      }

      if(_dataClientSock){
        if(_dataClientSock->connected()){
          sendByeMessage(_dataClientSock);
          _dataClientSockConnected = false;
          _dataClientSock->close();
        }
        delete _dataClientSock;
        _dataClientSock = NULL;
      }
    }

    void DataClientProxy::rebuildDataClientSocket(void){
      _dataClientSockConnected = rebuildClientConnection(_dataClientSock, _connectionString, ZMQ_DEALER, _clientIdentity, 0, NODE_MSG_READY);
    }

    void DataClientProxy::processControlMessage(zmsg& msg){
      if(msg.address() == NODE_MSG_HEARTBEAT){
        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED HEARTBEAT from server" << flush;

        //!<Register heartbit time mark
        _heartbeatedAt = s_clock();

        //!<Move heartbit time forward
        //_heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "ERROR: wrong control message received" << endl << msg.dump() << flush;
      }
    }

    void DataClientProxy::processDataMessage(zmsg& msg){
      _statRequests++;

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM dataclient" << endl << msg.dump() << flush;
      logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from upper server, n=" << _statRequests << flush;

      //!<Proxy message to inproc handler
      msg.send(*_inprocProxySock);

      logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT to inproc_proxy socket" << flush;

      //TODO: More detailed study the optimization for heartbeat from client side
      //!<Register heartbit time mark
      //_heartbeatedAt = s_clock();
      //!<Move heartbit time forward
      //_heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;
    }

    void DataClientProxy::handleExternalMessage(void){
      //!<Fetch message from dataclient
      zmsg clientMsg(*_dataClientSock);

      if(clientMsg.parts() == NODE_MSG_CMD_MIN_FIELDS){
        //!<Control message
        processControlMessage(clientMsg);
      }else{
        //!<Data message
        processDataMessage(clientMsg);
      }

      //!<Register heartbit time mark
      _heartbeatedAt = s_clock();
    }

    void DataClientProxy::handleInternalMessage(void){
      //!<Fetch message from inproc_reducer_out
      zmsg msg(*_inprocReducerOutSock);

      //!<Send response message to server
      msg.send(*_dataClientSock);

      _statResponses++;

      //TODO: use move time of heartbit when data message will be extended with route and resource usage information
      //!<Move heartbit time forward
      _heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_reducer_out" << endl << msg.dump() << flush;
      logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to dataclient socket from proxy node to upper server, n= " << _statResponses << " _heartbeatAt = '"
          << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
    }

    void DataClientProxy::heartbit(void){
      if(s_clock() > _heartbeatAt && _dataClientSockConnected){
        s_send(*_dataClientSock, NODE_MSG_HEARTBEAT);

        //!<Reset heartbit time
        _heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;

        logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << _clientIdentity << "] to server _heartbeatAt = '"
            << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
      }
    }

    void DataClientProxy::externalConnectionRefresh(void){
      if(((s_clock() - _heartbeatedAt) > (HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS)) && getInProgress() && _dataClientSockConnected){
        logger.log(NODE_LOG_MODE_DEBUG) << "RECONNECT client [" << _clientIdentity << "] because HEARTBIT not received during time of "
            << (HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS) << " msec" << flush;

        //!<Rebuild data socket
        rebuildDataClientSocket();

        //Reset time mark
        _heartbeatedAt = s_clock();

        //!<Move heartbit time forward
        _heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;
        logger.log(NODE_LOG_MODE_DEBUG) << " _heartbeatAt = '"
            << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
      }
    }

    std::string DataClientProxy::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << _statRequests << ",responses=" << _statResponses;
        ret = result.str();
      }else if(command == NODE_MSG_REBUILD_CLIENT_CONNECTION){
        //!<Rebuild client connection
        rebuildDataClientSocket();
      }else if(command == NODE_MSG_DISCONNECT_CLIENT_CONNECTION){
        //!<Disconnect client connection
        _dataClientSock->close();
        _dataClientSockConnected = false;
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "requests=" << _statRequests << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << _statResponses
            << NODE_ADMIN_COMMAND_DELIMITER << "clientIdentity=" << _clientIdentity << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatAt=" << static_cast<uint64_t>(_heartbeatAt)
            << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatedAt=" << static_cast<uint64_t>(_heartbeatedAt);

        ret = result.str();
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void* DataClientProxy::process(void){
      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      while(getInProgress()){
        try {
          //!<Init polling pool
          std::vector<zmq::pollitem_t> items;
          if(_dataClientSockConnected){
            items ={{*_dataClientSock, 0, ZMQ_POLLIN, 0},{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}}; //add new socket
          } else{
            items ={{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}}; //add new socket inproc_proxy
          }

          //!<Poll dataclient sockets
          zmq::poll(&items[0], items.size(), POLL_TIMEOUT);

          if(_dataClientSockConnected){
            //!<Handle dataclient socket's message
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_reducer_out socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleAdminMessage();
          }

          //!<Send heartbeat to server
          heartbit();

          //!<Heartbited interval check and connection refresh
          externalConnectionRefresh();
        }catch(zmq::error_t& error){
          std::cout << "DataClientProxy E: " << error.what() << std::endl;
        }
      }
      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }

  }
}
