#include <stdlib.h>
#include <time.h>
#include <random>

#include "DataServerProxy.hpp"

//DRCE FO
#include "DRCEFunctionalObject.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCEError.hpp"
//Sphinx FO
#include "SphinxFunctionalObject.hpp"
#include "SphinxMessageConst.hpp"
#include "SphinxError.hpp"
//Message cover object
#include "JsonMessageCover.hpp"

namespace HCE {
  namespace handlers {

    DataServerProxy::DataServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString, unsigned char nodeMode, int64_t startedAt,
        unsigned char logPriority) :
        Handler(name, ctx, connectionString, startedAt, logPriority), _clientIdentity(clientIdentity), _nodeMode(nodeMode), _previousClientNumber(0){

      //!<Init all the rest
      initialize();
    }

    DataServerProxy::~DataServerProxy(){
      deinitialize();
    }

    int DataServerProxy::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << _clientIdentity << "] with connection [" << _connectionString << "]" << flush;

      //!<Init sockets
      _inprocProxySock = NULL;
      _inprocReducerInSock = NULL;
      _dataServerSock = NULL;

      //!<Heartbeat at time mark
      _heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;

      //!<Send out periodical stats at regular intervals
      _pstatsAt = s_clock() + STAT_INTERVAL;

      //!<Stat fields
      _statClientRequests = 0;
      _statClientResponses = 0;
      _statRequests = 0;
      _statClientRequestsSize = 0;
      _statClientResponsesSize = 0;
      _statRequestsSize = 0;
      _periodicStatClientRequests = 0;
      _periodicStatClientResponses = 0;
      _statRequestsProcessed = 0;
      _statRequestsProcessedSize = 0;
      _periodicStatClientRequestsSize = 0;
      _periodicStatClientResponsesSize = 0;
      _statQPSClient = 0;
      _statQPSServer = 0;
      _statRPSClient = 0;
      _statRPSServer = 0;
      _statRequestsBandwidthClient = 0;
      _statRequestsBandwidthServer = 0;
      _statResponsesBandwidthClient = 0;
      _statResponsesBandwidthServer = 0;

      _clientsManager = new ClientsQueueManager(&_clientsQueue, 0);

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildServerConnection(_inprocProxySock, NODE_CS_INPROC_PROXY, ZMQ_DEALER);
      //!<Rebuild inproc_reducer_in connection
      rebuildClientConnection(_inprocReducerInSock, NODE_CS_INPROC_REDUCER_IN, ZMQ_DEALER);
      //!<Rebuild external server connection
      rebuildServerConnection(_dataServerSock, _connectionString, ZMQ_ROUTER);

      return true;
    }

    void DataServerProxy::deinitialize(void){
      if(_inprocProxySock){
        delete _inprocProxySock;
      }
      if(_inprocReducerInSock){
        delete _inprocReducerInSock;
      }
      if(_dataServerSock){
        delete _dataServerSock;
      }
      _inprocProxySock = NULL;
      _inprocReducerInSock = NULL;
      _dataServerSock = NULL;
      _clientsQueue.clear();
      if(_clientsManager){
        delete _clientsManager;
      }
    }

    void DataServerProxy::handleInternalMessage(void){
      //!<Fetch message from inprocontrol
      zmsg msg(*_inprocProxySock);
      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_proxy" << endl << msg.dump() << flush;

      //!<Create identity, messageId and body of dataclient's message
      std::string clientIdentity((const char*)msg.pop_front().c_str());
      std::string messageId((const char*)msg.pop_front().c_str());
      std::string messageBody((const char*)msg.pop_front().c_str());

      logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from upper server [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
          << " bytes, n=" << ++_statRequests << flush;

      if(_clientsQueue.empty()){
        //!<No clients connected, make response message and send response for reducer
        messageBody = makeEmptyResponseMessage(messageBody, EmptyResponseType::ertNoClients);
        //!<Insert message Id back
        msg.append(messageId.c_str());
        //!<Insert body back
        msg.append(messageBody.c_str());
        //!<Create current clients number
        std::stringstream clientsNumberStream;
        clientsNumberStream << _clientsQueue.size();
        //!<Insert clients number field
        msg.append(clientsNumberStream.str().c_str());
        //!<Send message
        msg.send(*_inprocReducerInSock);

        logger.log(NODE_LOG_MODE_TRACE) << "No client`s connection registered!" << endl << msg.dump() << flush;
      }else{

        logger.log(NODE_LOG_MODE_DEBUG) << "DataServerProxy::handleInternalMessage  _nodeMode" << _nodeMode << flush;
        //!<Some client(s) registered
        if(_nodeMode == NODE_MODE_PROXY || _nodeMode == NODE_MODE_ROUTER){
          //!<Re-send data request for all clients connected to dataserver
          for(std::vector<ClientWorkerItem>::iterator it = _clientsQueue.begin(); it < _clientsQueue.end(); ++it){
            //!<Create response message and fill with fields
            zmsg clientMsg(_clientIdentity.c_str());
            clientMsg.wrap(it->identity.c_str(), NULL);
            clientMsg.append(messageId.c_str());
            clientMsg.append(messageBody.c_str());

            //!<Send response
            clientMsg.send(*_dataServerSock);

            _statClientRequests++;
            _statClientRequestsSize += messageBody.size();

            logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT_ALL to dataserver socket for client [" << it->identity << "], message [" << messageId << "] body "
                << messageBody.size() << ", n=" << _statClientRequests << flush;
          }
        }else{
          //!<Send data request round-robin way for one client with low frequency of usage
          if(_nodeMode == NODE_MODE_SHARD || _nodeMode == NODE_MODE_SHARD_RND){

            logger.log(NODE_LOG_MODE_DEBUG) << "DataServerProxy::handleInternalMessage  (_nodeMode == NODE_MODE_SHARD || _nodeMode == NODE_MODE_SHARD_RND)" << flush;

            std::string identity;
            if (_nodeMode == NODE_MODE_SHARD_RND)
            {
              logger.log(NODE_LOG_MODE_DEBUG) << "DataServerProxy::handleInternalMessage  (_nodeMode == NODE_MODE_SHARD_RND)" << flush;
              //!<Initialize random seed
/*              std::random_device rd;
              std::mt19937_64 gen(rd());
              std::uniform_real_distribution<double> dis(0, _clientsQueue.size());
              //!<Generate index number limited clients queue size
              double uniformOn01 = dis(gen);
              size_t queueIndex = static_cast<size_t>(uniformOn01);*/

              logger.log(NODE_LOG_MODE_DEBUG) << "NODE_MODE_SHARD_RND  clientQueue.size() = " << _clientsQueue.size() << " previousClientNumber = " << _previousClientNumber << flush;

              size_t queueIndex = 0;
              do{
                queueIndex = getRandomNodeNumber(_clientsQueue.size());

                logger.log(NODE_LOG_MODE_DEBUG) << "NODE_MODE_SHARD_RND  getRandomNodeNumber() return queueIndex = " << queueIndex << flush;

                if (_clientsQueue.size()==1)
                {
                  logger.log(NODE_LOG_MODE_DEBUG) << "NODE_MODE_SHARD_RND  do-while broken because: _clientsQueue.size()==1" << flush;
                  break;
                }

              }while(queueIndex==_previousClientNumber);
              _previousClientNumber = queueIndex;
              //!<Inc freq of random element
              _clientsQueue[queueIndex].frequency++;
              //!<Save identity for next use
              identity = _clientsQueue[queueIndex].identity;

              logger.log(NODE_LOG_MODE_DEBUG) << "_RR identity=" << identity << ", queueIndex=" << queueIndex << ", _clientsQueue.size()= " << _clientsQueue.size() << flush;
            }
            else
            {
              logger.log(NODE_LOG_MODE_DEBUG) << "DataServerProxy::handleInternalMessage  (_nodeMode != NODE_MODE_SHARD_RND)" << flush;

              //!<Sort clients by frequency of queries done
              std::sort(_clientsQueue.begin(), _clientsQueue.end(), ClientWorkerItem::sortByFrequencyPredicate);
              //!<Inc freq of best (with minimal freq)
              _clientsQueue[0].frequency++;
              //!<Save identity for next use
              identity = _clientsQueue[0].identity;
            }
            //!<Create response message and fill with fields
            zmsg clientMsg(_clientIdentity.c_str());
            clientMsg.wrap(identity.c_str(), NULL);
            clientMsg.append(messageId.c_str());
            clientMsg.append(messageBody.c_str());

            //!<Send response
            clientMsg.send(*_dataServerSock);

            _statClientRequests++;
            _statClientRequestsSize += messageBody.size();

            logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT_RR to _dataServerSock socket for client " << identity.c_str() << " from [" << _clientIdentity.c_str() << "], message ["
                << messageId << "] body " << messageBody.size() << ", n=" << _statClientRequests << flush;
          }else{
            //!<Node mode not supported, make response message and send response for reducer
            messageBody = makeEmptyResponseMessage(messageBody, EmptyResponseType::ertBadNodeMode);
            //!<Insert message Id back
            msg.append(messageId.c_str());
            //!<Insert body back
            msg.append(messageBody.c_str());
            //!<Create current clients number
            std::stringstream clientsNumberStream;
            clientsNumberStream << _clientsQueue.size();
            //!<Insert clients number field
            msg.append(clientsNumberStream.str().c_str());
            //!<Send message
            msg.send(*_inprocReducerInSock);

            logger.log(NODE_LOG_MODE_TRACE) << "Node mode value " << _nodeMode << " not supported, message skipped" << endl << msg.dump() << flush;
          }
        }
      }
    }

    size_t DataServerProxy::getRandomNodeNumber(size_t maxNumber){
      //!<Initialize random seed
      std::random_device rd;
      std::mt19937_64 gen(rd());
      std::uniform_real_distribution<double> dis(0, maxNumber);
      //!<Generate index number limited clients queue size
      return static_cast<size_t>(dis(gen));
    }

    void DataServerProxy::registerClient(const std::string& identity){
      _clientsManager->remove(identity);
      _clientsManager->append(identity, false);

      logger.log(NODE_LOG_MODE_DEBUG) << "READY message from client [" << identity << "] received" << flush;
    }

    //TODO: experimental_heartbitResponse
    /*
     void DataServerProxy::heartbitResponse(const std::string& identity){
     //!<Send response heartbeat
     zmsg hb_msg(NODE_MSG_HEARTBEAT.c_str());
     hb_msg.wrap(identity.c_str(), NULL);
     hb_msg.send(*_dataServerSock);

     log << "HEARTBEAT message for client [" << identity << "] sent as response";
     logFlush();
     }
     */

    void DataServerProxy::processControlMessage(zmsg& msg){
      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());

      char* adrp = msg.address();
      std::string addr((adrp == 0) ? "" : adrp);

      if(addr == NODE_MSG_READY){
        registerClient(clientIdentity);
      }else if(addr == NODE_MSG_HEARTBEAT){
        //!<Refresh client state in queue
        _clientsManager->refresh(clientIdentity);

        logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message from client [" << clientIdentity << "] received" << flush;

        //!<Send heartbit response
        //heartbitResponse(clientIdentity); //TODO: experimental_heartbitResponse
      }else if(addr == NODE_MSG_BYE){
        //!<Delete client from queue
        _clientsManager->remove(clientIdentity);

        logger.log(NODE_LOG_MODE_DEBUG) << "BYE message from client [" << clientIdentity << "] received" << flush;
      }else{
        //!<Invalid message type
        logger.log(NODE_LOG_MODE_TRACE) << "INVALID command message from client [" << clientIdentity << "]" << endl << msg.dump() << flush;
      }
    }

    void DataServerProxy::processDataMessage(zmsg& msg){
      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());

      if(!_clientsManager->refresh(clientIdentity)){
        logger.log(NODE_LOG_MODE_TRACE) << "MSG STRUCTURE ERROR wrong client identity [" << clientIdentity << "]"
            << endl << "Clients:[" << _clientsManager->getDump() << "]" << endl << msg.dump() << flush;
      }

      //!<Get message Id
      std::string messageId((const char*)msg.pop_front().c_str());
      //!<Get message body
      std::string messageBody((const char*)msg.pop_front().c_str());
      _statClientResponsesSize += messageBody.size();
      logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED RESPONSE from down client [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
          << " bytes n=" << _statClientResponses << flush;

      //!<Insert message Id back
      msg.append(messageId.c_str());
      //!<Insert body back
      msg.append(messageBody.c_str());
      //!<Create current clients number
      std::stringstream clientsNumberStream;
      clientsNumberStream << _clientsQueue.size();
      //!<Insert clients number field
      msg.append(clientsNumberStream.str().c_str());

      //!<Send message
      msg.send(*_inprocReducerInSock);

      logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to inproc_reducer_in socket, message [" << messageId << "] body " << messageBody.size() << " bytes sent" << flush;
    }

    std::string DataServerProxy::makeEmptyResponseMessage(const std::string& messageBody, EmptyResponseType responseType){
      HCE::JsonMessageCover msgCover(messageBody);
      std::string resultMessageBody;
      switch(msgCover.getType()){
        case HCE::types::MessageType::mtSphinx: {

            if (responseType==EmptyResponseType::ertNoClients){
              resultMessageBody = HCE::sphinx::SphinxFunctionalObject::makeErrorMessage(msgCover.getData(),
                  HCE::sphinx::ERROR_TARGET_CLIENTS_NOT_AVAILABLE,
                  HCE::sphinx::message_const::targetClientsNotAvailable);
            }else if(responseType==EmptyResponseType::ertBadNodeMode){
              resultMessageBody = HCE::sphinx::SphinxFunctionalObject::makeErrorMessage(msgCover.getData(),
                   HCE::sphinx::ERROR_NODE_NOT_SUPPORTED,
                   HCE::sphinx::message_const::nodeNotSupported);
            }
          }
          break;
        case HCE::types::MessageType::mtDrce: {

            if (responseType==EmptyResponseType::ertNoClients){
              resultMessageBody = HCE::drce::DRCEFunctionalObject::makeErrorMessage(msgCover.getData(),
                  HCE::drce::DRCETaskRequest::TaskState::NOT_SET_AS_NEW,
                  HCE::drce::ERROR_TARGET_CLIENTS_NOT_AVAILABLE,
                  HCE::drce::message_const::targetClientsNotAvailable);
            }
            else if(responseType==EmptyResponseType::ertBadNodeMode){
              resultMessageBody = HCE::drce::DRCEFunctionalObject::makeErrorMessage(msgCover.getData(),
                  HCE::drce::DRCETaskRequest::TaskState::NOT_SET_AS_NEW,
                  HCE::drce::ERROR_NODE_NOT_SUPPORTED,
                  HCE::drce::message_const::nodeNotSupported);
            }
          }
          break;
        default:;
      }

      std::string jsonMsg;
      msgCover.setData(resultMessageBody);
      if(!msgCover.serialize(jsonMsg)){
        logger.log(NODE_LOG_MODE_ERROR) << "Serialization error of cover object from empty results " << msgCover.getErrorCode() << " : " << msgCover.getErrorMsg() << flush;
      }

      return jsonMsg;
    }

    void DataServerProxy::handleExternalMessage(void){
      //!<Fetch message from dataserver
      zmsg msg(*_dataServerSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM dataserver" << endl << msg.dump() << flush;

      //!<Process dataserver's message
      if(msg.parts() == NODE_MSG_DATA_MIN_FIELDS){
        //!<Control message
        processControlMessage(msg);
      }else if(msg.parts() > NODE_MSG_DATA_MIN_FIELDS){
        //!<Inc data responses counter
        _statClientResponses++;

        //!<Data message
        processDataMessage(msg);
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "ERROR of message structure, wrong fields number " << msg.parts() << endl << msg.dump() << flush;
      }
    }

    std::string DataServerProxy::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "clientRequests=" << _statClientRequests << ",clientResponses=" << _statClientResponses << ",clientRequestsSize="
            << _statClientRequestsSize << ",clientResponsesSize=" << _statClientResponsesSize << ",Requests=" << _statRequests << ",requestsSize=" << _statRequestsSize;
        ret = result.str();
      }else if(command == NODE_MSG_rebuildServerConnection){
        //!<Rebuild server connection
        rebuildServerConnection(_dataServerSock, _connectionString, ZMQ_ROUTER);
      }else if(command == NODE_MSG_DISCONNECT_SERVER_CONNECTION){
        //!<Disconnect server connection
        _dataServerSock->close();
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "clientRequests=" << _statClientRequests << NODE_ADMIN_COMMAND_DELIMITER << "clientResponses="
            << _statClientResponses << NODE_ADMIN_COMMAND_DELIMITER << "clientRequestsSize=" << _statClientRequestsSize << NODE_ADMIN_COMMAND_DELIMITER << "clientResponsesSize="
            << _statClientResponsesSize << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << _statRequests << NODE_ADMIN_COMMAND_DELIMITER << "requestsSize=" << _statRequestsSize
            << NODE_ADMIN_COMMAND_DELIMITER << "nodeMode=" << static_cast<int>(_nodeMode) << NODE_ADMIN_COMMAND_DELIMITER << "clients=" << _clientsManager->getDump()
            << NODE_ADMIN_COMMAND_DELIMITER << "statQPSClient=" << _statQPSClient << NODE_ADMIN_COMMAND_DELIMITER << "statQPSServer=" << _statQPSServer << NODE_ADMIN_COMMAND_DELIMITER
            << "statRPSClient=" << _statRPSClient << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidthClient=" << _statRequestsBandwidthClient << NODE_ADMIN_COMMAND_DELIMITER
            << "statRequestsBandwidthServer=" << _statRequestsBandwidthServer << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidthClient=" << _statResponsesBandwidthClient;

        ret = result.str();
      }else if(command == NODE_MSG_MANAGER_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_nodeMode));
      }else if(command == NODE_MSG_MANAGER_MODE_SET){
        //!<Parse command parameters and fetch log level value
        int nodeMode = _nodeMode;
        //Parse log priority from parameters
        std::istringstream(parameters) >> nodeMode;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_nodeMode));
        //Set new log priority
        _nodeMode = nodeMode;
      }else if(command == NODE_MSG_MANAGER_PURGE_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_clientsManager->getPurgeMode()));
      }else if(command == NODE_MSG_MANAGER_PURGE_MODE_SET){
        //!<Parse command parameters
        int purgeMode = 0;
        //Parse log priority from parameters
        std::istringstream(parameters) >> purgeMode;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_clientsManager->getPurgeMode()));
        //Set new log priority
        _clientsManager->setPurgeMode(purgeMode);
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void DataServerProxy::logPeriodicStatistics(void){
      //!<Out periodical stats
      if(s_clock() > _pstatsAt){

        _statQPSClient = (_statClientRequests - _periodicStatClientRequests) / (float)((STAT_INTERVAL) / 1000.0);
        _statQPSServer = (_statRequests - _statRequestsProcessed) / (float)((STAT_INTERVAL) / 1000.0);
        _statRPSClient = (_statClientResponses - _periodicStatClientResponses) / (float)((STAT_INTERVAL) / 1000.0);
        _statRPSServer = 0;
        _statRequestsBandwidthClient = (_statClientRequestsSize - _periodicStatClientRequestsSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statRequestsBandwidthServer = (_statRequestsSize - _statRequestsProcessedSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statResponsesBandwidthClient = (_statClientResponsesSize - _periodicStatClientResponsesSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statResponsesBandwidthServer = 0;

        _statRequestsProcessed = _statRequests;
        _statRequestsProcessedSize = _statRequestsSize;
        _periodicStatClientRequests = _statClientRequests;
        _periodicStatClientResponses = _statClientResponses;
        _periodicStatClientRequestsSize = _statClientRequestsSize;
        _periodicStatClientResponsesSize = _statClientResponsesSize;

        logger.log(NODE_LOG_MODE_DEBUG) << "SRequests=" << _statRequests << ", CRequests=" << _statClientRequests << ", CResponses=" << _statClientResponses
            << ", SQPS=" << _statQPSServer << ", CQPS=" << _statQPSClient << ", CRPS=" << _statRPSClient << ", SRequestsSize=" << _statRequestsSize
            << ", CRequestsSize=" << _statClientRequestsSize << ", CResponsesSize=" << _statClientResponsesSize
            << ", SRequestsBandwidth=" << _statRequestsBandwidthServer << ", CRequestsBandwidth=" << _statRequestsBandwidthClient
            << ", CResponsesBandwidth=" << _statResponsesBandwidthClient << flush;

        //Shift next log stat action time
        _pstatsAt = s_clock() + STAT_INTERVAL;
      }
    }

    void DataServerProxy::heartbit(void){
      //!<Send heartbeats to idle clients if it's time
      if(s_clock() > _heartbeatAt){
        for(std::vector<ClientWorkerItem>::iterator it = _clientsQueue.begin(); it < _clientsQueue.end(); ++it){
          //if(s_clock() > it->expiry){//TODO: experimental
          zmsg msg(NODE_MSG_HEARTBEAT.c_str());
          msg.wrap(it->identity.c_str(), NULL);
          msg.send(*_dataServerSock);

          logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message for client [" << it->identity.c_str() << "] sent by timer reason" << flush;
          //}
        }

        //Shift next heartbit send time
        _heartbeatAt = s_clock() + HEARTBEAT_INTERVAL;
      }
    }

    void* DataServerProxy::process(void){
      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      while(getInProgress()){
        try {
          std::vector<zmq::pollitem_t> items;
          if(_dataServerSock->connected()){
            items ={{*_dataServerSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          } else{
            items ={{*_inprocProxySock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          }

          //!<Poll dataserver
          zmq::poll(&items[0], items.size(), POLL_TIMEOUT);

          //!<Handle dataserver socket's message if socket is connected
          if(_dataServerSock->connected()){
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_proxy socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleAdminMessage();
          }

          //!<Heartbeats for clients
          heartbit();

          //!<Log stats
          logPeriodicStatistics();

          //!<Remove expired clients from queue
          std::string r = _clientsManager->purge();
          if(r != ""){
            logger.log(NODE_LOG_MODE_DEBUG) << "PURGED_CLIENTS: " << r << flush;
          }
        }catch(zmq::error_t& error){
          std::cout << "DataServerProxy E: " << error.what() << std::endl;
        }
      }

      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }

  }
}
