/**
 * @file DRCEResourceExtractor.hpp
 *
 * Implementation classes DiskSpaceExtractor, CpuUsageExtractor, MemoryExtractor and ProcessInfoExtractor
 * of a Distributed Remote Command Execution functional object in HCE cluster node.
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef DRCE_RESOURCE_EXTRACTOR_HPP
#define DRCE_RESOURCE_EXTRACTOR_HPP

#include <iostream>
#include <utility>
#include <Poco/Exception.h>
#include <Poco/Mutex.h>
#include <Poco/ScopedLock.h>

#include "DRCEResourceUsage.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
class DiskSpaceExtractor
{
public:
  DiskSpaceExtractor(void);
  ~DiskSpaceExtractor(void) {}

  void update(void) throw (Poco::Exception);

  void setDiskSize(size_t diskSize_);

  size_t getDiskSize(void) const {return diskSize;}
  size_t getDiskUsed(void) const {return diskUsed;}
  size_t getDiskFree(void) const {return diskFree;}
  double getDiskUsedPercent(void) const {return diskUsedPercent;}
  double getDiskFreePercent(void) const {return diskFreePercent;}
private:
  size_t diskSize;
  size_t diskUsed;
  size_t diskFree;
  double diskUsedPercent;
  double diskFreePercent;

  size_t externalDiskSize;
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
class CpuUsageExtractor
{
public:/*
  struct Pstat
  {
    Pstat(void);
    Pstat(const Pstat& rhs);
    Pstat(Pstat&& rhs);
    Pstat& operator=(const Pstat& rhs);
    Pstat& operator=(Pstat&& rhs);

    void clear(void);

    Pstat& operator+=(const Pstat& rhs);

    long unsigned int utimeTicks;
    long int cutimeTicks;
    long unsigned int stimeTicks;
    long int cstimeTicks;
    long unsigned int vsize;  // virtual memory size in bytes
    long unsigned int rss;    // resident set size in bytes

    long unsigned int cpuTotalTime;
  };*/
//  static int getUsage(const pid_t pid, struct Pstat* result);
/*  static void calcCpuUsagePct(const struct Pstat* curUsage,
                          const struct Pstat* lastUsage,
                          double* ucpuUsage, double* scpuUsage);
  static void calcCpuUsage(const struct Pstat* curUsage,
                      const struct Pstat* lastUsage,
                      long unsigned int* ucpuUsage,
                      long unsigned int* scpuUsage);
*/
  static double getCpuUsage(const std::vector<pid_t>& ids);

  CpuUsageExtractor(void);
  ~CpuUsageExtractor(void) {}

  void update(void);

  double getCpuUsage(void) const {return cpuUsage;}
  double getIOWait(void) const {return iowait;}

private:
  double cpuUsage;
  unsigned long long lastBusy;
  unsigned long long lastWork;

  double iowait;
  unsigned long long lastIOWait;
  unsigned long long lastSum;

  static void getCpuData(pid_t pid, std::pair<unsigned long int, unsigned long int>& p);
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
class MemoryExtractor
{
public:
  struct MemInfo
  {
    MemInfo(void);
    MemInfo(MemInfo&& rhs);
    MemInfo& operator=(MemInfo&& rhs);

    size_t memTotal;
    size_t memFree;
    size_t swapTotal;
    size_t swapFree;
    size_t vmallocTotal;
    size_t vmallocUsed;
  };

  MemoryExtractor(void);
  ~MemoryExtractor(void) {}

  void update(void);

  size_t getVramUsed(void) const {return vramUsed;}
  size_t getRramUsed(void) const {return rramUsed;}
  size_t getVramFree(void) const {return vramFree;}
  size_t getRramFree(void) const {return rramFree;}

  double getVramUsedPercent(void) const {return vramUsedPercent;}
  double getRramUsedPercent(void) const {return rramUsedPercent;}
  double getVramFreePercent(void) const {return vramFreePercent;}
  double getRramFreePercent(void) const {return rramFreePercent;}

  size_t getVramTotal(void) const {return vramTotal;}
  size_t getRramTotal(void) const {return rramTotal;}

  void setVramTotal(size_t vramTotal_);
  void setRramTotal(size_t rramTotal_);
private:
  static MemInfo getMemoryInfo(void);
private:
  size_t vramUsed;
  size_t rramUsed;
  size_t vramFree;
  size_t rramFree;
  double vramUsedPercent;
  double rramUsedPercent;
  double vramFreePercent;
  double rramFreePercent;

  size_t vramTotal;
  size_t rramTotal;

  size_t externalVramTotal;
  size_t externalRramTotal;
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
class AsyncTasks;
class ProcessInfoExtractor
{
public:
  struct FieldsArray
  {
    FieldsArray(void);
    FieldsArray(const FieldsArray& rhs);
    FieldsArray(FieldsArray&& rhs);

    FieldsArray& operator=(const FieldsArray& rhs);
    FieldsArray& operator=(FieldsArray&& rhs);

    double cpu;
    size_t vram;
    size_t rram;
    size_t threads;
  };

  ProcessInfoExtractor(void) {};
  ~ProcessInfoExtractor(void) {};

  static FieldsArray extract(AsyncTasks& asyncTask);
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
class ResourceUsageExtractor
{
public:
  static ResourceUsageExtractor& getInstance(void);
  static void shutdown(void);

  void update(void);
  ResourceUsage extract(void) throw (Poco::Exception);

  CpuUsageExtractor& getCpuUsageExtractor(void);
  DiskSpaceExtractor& getDiskSpaceExtractor(void);
  MemoryExtractor& getMemoryExtractor(void);

  void setMaxThreadsCount(size_t maxThreadsCount_);
  size_t getMaxThreadsCount(void);

  void setMaxProcessesCount(size_t maxProcessesCount_);
  size_t getMaxProcessesCount(void);

  void setMaxDiskSize(size_t maxDiskSize_);
  size_t getMaxDiskSize(void);

  void setMaxVramSize(size_t maxVramSize_);
  size_t getMaxVramSize(void);

  void setMaxRramSize(size_t maxRramSize_);
  size_t getMaxRramSize(void);

private:
  unsigned int getMaxProcesses(void); //temporary implementation, possible in future it can be remove
  unsigned int getMaxThreads(void);   //temporary implementation, possible in future it can be remove
  unsigned int extractValue(const std::string& fileName); //temporary implementation, possible in future it can be remove
protected:
  static ResourceUsageExtractor* pHandler;
  static Poco::Mutex mutex;

  CpuUsageExtractor cpuUsageExtractor;
  DiskSpaceExtractor diskSpaceExtractor;
  MemoryExtractor memoryExtractor;

  size_t maxThreadsCount;
  size_t maxProcessesCount;
private:
  ResourceUsageExtractor(void);
  ~ResourceUsageExtractor(void) {};
  ResourceUsageExtractor(const ResourceUsageExtractor&);
  ResourceUsageExtractor& operator=(const ResourceUsageExtractor&);
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce 
} // end namespace HCE 

#endif // DRCE_RESOURCE_EXTRACTOR_HPP
