/**
 * @file DRCEResourceMonitor.hpp
 *
 * Implementation class DRCEResourceMonitor of a Distributed Remote Command Execution functional object in HCE cluster node.
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef DRCE_RESOURCE_MONITOR_HPP
#define DRCE_RESOURCE_MONITOR_HPP

#include <Poco/Thread.h>
#include <Poco/Mutex.h>
#include <string>
#include <map>

#include "DRCEAsyncTasksQueue.hpp"
#include "DRCEStatVariables.hpp"
#include "LoggerStream.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
class DRCEResourceMonitor : Poco::Runnable
{
  enum {DEFAULT_TIME_PERIOD=10000};
public:
  DRCEResourceMonitor(DRCEAsyncTasksQueue& asyncTasksQueue_,
      CustomMessage& message_);
  ~DRCEResourceMonitor(void);

  void run(void);

  void terminate(void);
  bool isTerminated(void);

  void setTimePeriod(unsigned int timePeriod_);
  unsigned int getTimePeriod(void);

  void fillFieldsArray(DRCEResultData& resultData);
  static void fillFieldsArray(DRCEResultData& resultData, DRCEAsyncTasksQueue& asyncTasksQueue_, CustomMessage& message_, DRCEResourceMonitor* pResourceMonitor=nullptr);
  static void fillFieldsArray(DRCEResultDataItem& resultDataItem, DRCEAsyncTasksQueue& asyncTasksQueue_, CustomMessage& message_, DRCEResourceMonitor* pResourceMonitor=nullptr);

  void setStatVariables(const StatVariables& rhs);
  void setStatVariables(StatVariables&& rhs);
  StatVariables getStatVariables(void);

  void addTimeAsyncTask(size_t value);
  void addTimeSyncTask(size_t value);
  void addSizeInputBufferSyncTasks(size_t value);
  void addSizeInputBufferAsyncTasks(size_t value);
  void addSizeOutputBufferSyncTasks(size_t value);
  void addSizeOutputBufferAsyncTasks(size_t value);
  void addCpuUsageAsyncTasks(double value);
  void addMemoryUsageAsyncTasks(size_t value);
  void addCountAsyncTasksForMinute(double value);
  void addCountSyncTasksForMinute(double value);
  void incrementCountSyncTasks(void);
  void incrementCountAsyncTasks(void);
  void incrementCountSyncTasksFail(void);
  void incrementCountAsyncTasksFail(void);
private:
  template <class T> void addStatValue(T value, StatVariableData<T>& statVariableData, size_t& count);
  void countTasksForMinute(Poco::Timestamp& ts, size_t& countSyncTasks, size_t& countAsyncTasks);
  StatVariables recountStatVariables(void);
private:
  Poco::Thread thread;
  Poco::Mutex mutex;
  volatile bool terminated;
  unsigned int timePeriod;

  DRCEAsyncTasksQueue& asyncTasksQueue;
  CustomMessage& message;
  StatVariables statVariables;
  StatVariables counterStatVariables;

  struct AvgCounts
  {
    size_t timeAsyncTaskCount = 0;
    size_t timeSyncTaskCount = 0;
    size_t sizeInputBufferSyncTasksCount = 0;
    size_t sizeOutputBufferSyncTasksCount = 0;
    size_t sizeInputBufferAsyncTasksCount = 0;
    size_t sizeOutputBufferAsyncTasksCount = 0;
    size_t cpuUsageAsyncTasksCount = 0;
    size_t memoryUsageAsyncTasksCount = 0;
    size_t countSyncTasksForMinuteCount = 0;
    size_t countAsyncTasksForMinuteCount = 0;
  } avgCounts;
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce
} // end namespace HCE

#endif // DRCE_RESOURCE_MONITOR_HPP
